from qtpy import QtWidgets, QtGui
from loguru import logger

import bw2data as bd
import pandas as pd

from activity_browser import app
from activity_browser.ui import widgets, delegates, core


class CalculationSetupsPane(widgets.ABAbstractPane):
    title = "Calculation Setups"
    unique = True

    def __init__(self, parent):
        """
        Initializes the CalculationSetupsPane.

        This constructor sets up the view and model for displaying calculation setups,
        configures the view's appearance and behavior, and builds the layout while
        connecting necessary signals.

        Args:
            parent (QtWidgets.QWidget): The parent widget for this pane.
        """
        super().__init__(parent)
        self.model = CalculationSetupsModel(parent=self)
        self.view = CalculationSetupsView()
        self.view.setModel(self.model)

        self.view.setAlternatingRowColors(True)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        """
        Connects the signals to the appropriate slots.
        """
        app.signals.meta.calculation_setups_changed.connect(self.sync)

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the model with the current state of the calculation setups.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")
        df = self.build_df()
        self.model.set_dataframe(df)
        self.view.resizeColumnToContents(0)

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the calculation setups.

        Returns:
            pd.DataFrame: The DataFrame containing the calculation setups data.
        """
        data = []
        for cs in bd.calculation_setups:
            data.append(
                {
                    "name": cs,
                    "functional_units": len(bd.calculation_setups[cs].get("inv", [])),
                    "impact_categories": len(bd.calculation_setups[cs].get("ia", [])),
                }
            )

        cols = ["name", "functional_units", "impact_categories"]

        return pd.DataFrame(data, columns=cols)


class CalculationSetupsView(widgets.ABTreeView):
    """
    A view that displays the calculation setups in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "name": delegates.StringDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(app.actions.CSNew),
            lambda m, p: m.add(app.actions.CSOpen, p.calculation_setups,
                                  enable=bool(p.calculation_setups)),
            lambda m, p: m.add(app.actions.CSDelete, p.calculation_setups,
                                  enable=bool(p.calculation_setups)),
            lambda m, p: m.add(app.actions.CSRename, p.calculation_setups[0] if p.single_selection else None,
                                  enable=p.single_selection),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.CSCalculate, p.calculation_setups[0] if p.single_selection else None,
                                  enable=p.single_selection),
        ]

    @property
    def calculation_setups(self):
        if not self.selectedIndexes():
            return []
        names = self.model().values_from_indices("name", self.selectedIndexes())
        return list(set(names))

    @property
    def single_selection(self):
        return len(self.calculation_setups) == 1

    class HeaderMenu(QtWidgets.QMenu):
        """
        A header menu for the DatabasesView. Currently not used.
        """

        def __init__(self, *args, **kwargs):
            super().__init__()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        """
        Handles the mouse double click event to open the selected calculation setups.

        Args:
            event (QtGui.QMouseEvent): The mouse double click event.
        """
        index = self.indexAt(event.pos())

        if not index.isValid():
            return

        row = self.model().row(index)
        if row is None:
            return

        app.actions.CSOpen.run(row["name"])


    def dragMoveEvent(self, event) -> None:
        pass

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
            for key in keys:
                act = bd.get_node(key=key)
                if act["type"] not in bd.labels.product_node_types + ["processwithreferenceproduct"]:
                    keys.remove(key)

            if not keys:
                return

            event.accept()

    def dropEvent(self, event) -> None:
        event.accept()

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        for key in keys:
            act = bd.get_node(key=key)
            if act["type"] not in bd.labels.product_node_types + ["processwithreferenceproduct"]:
                keys.remove(key)

        functional_units = [{key: 1.0} for key in keys]

        app.actions.CSNew.run(functional_units=functional_units)


class CalculationSetupsModel(core.ABTreeModel):
    """
    A model representing the data for the calculation setups.
    """

    def fontData(self, index):
        """
        Provides font data for the model.

        Args:
            index: The index for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the index.
        """
        column_name = self.column_name(index)

        if column_name == "name":
            font = QtGui.QFont()
            font.setWeight(QtGui.QFont.Weight.DemiBold)
            return font

        return None




from activity_browser import app
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons

import bw2data as bd

from .project.project_switch import ProjectSwitch


class MetaDataStoreCacheClear(ABAction):

    icon = qicons.right
    text = "Clear Metadata Store Cache"
    tool_tip = "Clear the Metadata Store cache and reload the current project"

    @staticmethod
    @exception_dialogs
    def run():
        app.metadata.clear_cache()
        ProjectSwitch.run(bd.projects.current, reload=True)



import threading

from loguru import logger
from time import time
from threading import Thread

import pandas as pd
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt, QModelIndex

import bw2data as bd

from activity_browser import ui, app
from activity_browser.ui import core, widgets, delegates, icons
from activity_browser.bwutils.commontasks import database_is_locked, database_is_legacy, is_node_biosphere, nodes_to_excel


NODETYPES = {
    "all_nodes": [],
    "processes": ["process", "multifunctional", "processwithreferenceproduct", "nonfunctional"],
    "products": ["product", "processwithreferenceproduct", "waste"],
    "biosphere": ["natural resource", "emission", "inventory indicator", "economic", "social"],
}


class DatabaseProductsPane(widgets.ABAbstractPane):
    """
    A widget that displays products related to a specific database.

    Attributes:
        database (bd.Database): The database to display products for.
        model (ProductModel): The model containing the data for the products.
        table_view (ProductView): The view displaying the products.
        search (widgets.ABLineEdit): The search bar for quick search.
    """
    def __init__(self, parent, db_name: str):
        """
        Initializes the DatabaseProductsPane widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
            db_name (str): The name of the database to display products for.
        """
        self.name = "database_products_pane_" + db_name

        super().__init__(parent)

        self.database = bd.Database(db_name)
        self.title = db_name
        self.simple = True

        # initialize the model
        self.model = ProductModel(parent=self, chunk_size=20, enable_sorting=True)

        # Create the QTableView and set the model
        self.table_view = ProductView(self, db_name=db_name)
        self.table_view.setUniformRowHeights(True)
        self.table_view.setModel(self.model)

        self.search_bar = widgets.MetaDataAutoCompleteTextEdit(self)
        self.search_bar.database_name = db_name
        self.search_bar.setMaximumHeight(30)
        self.search_bar.setPlaceholderText("Quick Search")

        # Create loading indicator with spinner
        self.loading_spinner = QtWidgets.QProgressBar()
        self.loading_spinner.setRange(0, 0)  # Indeterminate/busy indicator
        self.loading_spinner.setTextVisible(False)
        self.loading_spinner.setMaximumWidth(200)
        self.loading_spinner.setMaximumHeight(20)
        
        self.loading_label = widgets.ABLabel("Loading database...")
        self.loading_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        font = self.loading_label.font()
        font.setPointSize(14)
        self.loading_label.setFont(font)
        self.loading_label.setStyleSheet("color: gray; padding: 10px;")

        # Create simple/detailed view toggle
        self.view_toggle = QtWidgets.QCheckBox("Details")
        self.view_toggle.setChecked(not self.simple)
        self.view_toggle.setToolTip("Toggle between simple and detailed view")

        self.build_layout()
        self.connect_signals()
        self.update_loading_state()
        self.sync()

    def build_layout(self):
        # Create a stacked layout to switch between loading and table view
        self.stacked_layout = QtWidgets.QStackedLayout()
        
        # Page 0: Loading indicator with spinner
        loading_widget = QtWidgets.QWidget(self)
        loading_layout = QtWidgets.QVBoxLayout(loading_widget)
        loading_layout.addStretch()
        loading_layout.addWidget(self.loading_spinner, alignment=Qt.AlignmentFlag.AlignCenter)
        loading_layout.addWidget(self.loading_label)
        loading_layout.addStretch()
        self.stacked_layout.addWidget(loading_widget)
        
        # Page 1: Table view
        table_widget = QtWidgets.QWidget(self)
        table_layout = QtWidgets.QVBoxLayout(table_widget)
        table_layout.setSpacing(0)
        table_layout.setContentsMargins(0, 0, 0, 0)
        table_layout.addWidget(self.table_view)
        self.stacked_layout.addWidget(table_widget)

        # Create top bar with search and toggle
        top_bar = QtWidgets.QHBoxLayout()
        top_bar.addWidget(self.search_bar)
        top_bar.addWidget(self.view_toggle)

        layout = QtWidgets.QVBoxLayout(self)
        layout.addLayout(top_bar)
        layout.addLayout(self.stacked_layout)

        # Set the table view as the central widget of the window
        self.setLayout(layout)

    def connect_signals(self):
        app.signals.metadata.synced.connect(self.on_metadata_changed)
        app.signals.database.deleted.connect(self.on_database_deleted)

        self.view_toggle.checkStateChanged.connect(self.on_mode_switch)
        self.search_bar.textChangedDebounce.connect(self.sync)

    def on_metadata_changed(self, added, updated, deleted):
        # Check if primary data has finished loading
        self.update_loading_state()
        
        if any(db == self.database.name for db, code in added | updated | deleted):
            self.sync()

    def update_loading_state(self):
        """
        Updates the loading state based on whether primary metadata has loaded.
        Shows the loading indicator if primary data is still loading, otherwise shows the table.
        """
        if app.metadata.loader.secondary_status == "done":
            # Show table view
            self.stacked_layout.setCurrentIndex(1)
        else:
            # Show loading indicator
            self.stacked_layout.setCurrentIndex(0)

    def sync(self):
        """
        Synchronizes the widget with the current state of the database.
        """
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        t = time()
        df = self.build_df()

        if self.search_bar.toPlainText().strip():
            # Reset sorting when searching
            self.model.sorted_column = None
            self.model.sort_order = Qt.SortOrder.AscendingOrder

        self.model.set_dataframe(df)

        self.update_table_style()
        self.update_column_visibility()

        logger.debug(f"Synced DatabaseProductsPane in {time() - t:.2f} seconds")

    def update_table_style(self):
        self.table_view.header().setHidden(self.simple)
        self.table_view.viewport().setBackgroundRole(
            QtGui.QPalette.ColorRole.Window if self.simple else QtGui.QPalette.ColorRole.Base)
        self.table_view.setFrameShape(
            QtWidgets.QFrame.Shape.NoFrame if self.simple else QtWidgets.QFrame.Shape.StyledPanel)

    def update_column_visibility(self):
        columns = self.model.columns()
        df = self.model.df

        for index, col in enumerate(columns):
            if col == "index":
                continue
            if col == "node":
                self.table_view.setColumnHidden(index, not self.simple)
                continue

            if df[col].isna().all() or self.simple:
                self.table_view.hideColumn(index)
            else:
                self.table_view.showColumn(index)

        self.table_view.reset()

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the database products.

        Returns:
            pd.DataFrame: The DataFrame containing the products data.
        """
        t = time()
        cols = ["name", "key", "processor", "product", "type", "unit", "location", "id", "categories", "properties"]

        query = self.search_bar.toPlainText().strip()
        if query:
            df = app.metadata.search_database(query, self.database.name, cols)
        else:
            df = app.metadata.get_database_metadata(self.database.name, cols)

        processors = set(df["processor"].dropna().unique())
        df = df.drop(processors, errors="ignore")
        df.rename(columns={"id": "_id"}, inplace=True)

        if not df.properties.isna().all():
            props_df = df[df.properties.notna()]
            props_df = pd.DataFrame(list(props_df.get("properties")), index=props_df.key)
            props_df.rename(lambda col: f"property_{col}", axis="columns", inplace=True)

            df = df.merge(
                props_df,
                left_on="key",
                right_index=True,
                how="left",
            )

        df["node"] = None

        cols = ["name", "product", "categories", "unit", "location", "key", "processor", "type", "node"]
        cols += [col for col in df.columns if col.startswith("property")]
        cols += ["_id"]

        logger.debug(f"Built DatabaseProductsPane dataframe in {time() - t:.2f} seconds")

        return df[cols].reset_index(drop=True)

    def on_database_deleted(self, db_name: str):
        """
        Handles the database deleted signal by closing the widget if the database is deleted.

        Args:
            db_name (str): The name of the deleted database.
        """
        if db_name == self.database.name:
            self.deleteLater()

    def on_mode_switch(self, check: Qt.CheckState):
        """
        Handles the mode switch between simple and detailed view.

        Args:
            check (Qt.CheckState): The check state of the toggle.
        """
        self.simple = check == Qt.CheckState.Unchecked
        self.update_table_style()
        self.update_column_visibility()


class ProductView(ui.widgets.ABTreeView):
    """
    A view that displays the products in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "categories": delegates.ListDelegate,
        "key": delegates.StringDelegate,
        "processor": delegates.StringDelegate,
        "node": delegates.CardDelegate,
    }

    class ContextMenu(ui.widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(app.actions.ActivityOpen, p.selected_activities,
                               text="Open process" if len(p.selected_activities) == 1 else "Open processes",
                               enable=len(p.selected_activities) > 0
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.ActivityNewProcess, p.db_name,
                               enable=not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(app.actions.ActivityDuplicate, p.selected_activities,
                               text="Duplicate process" if len(p.selected_activities) == 1 else "Duplicate processes",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(app.actions.ActivityDuplicateToDB, p.selected_activities,
                               text="Duplicate process to database" if len(p.selected_activities) == 1 else "Duplicate processes to database",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.ActivityDelete, p.selected_activities,
                               text="Delete process" if len(p.selected_activities) == 1 else "Delete processes",
                               enable=len(p.selected_activities) > 0 and not database_is_locked(p.db_name),
                               ),
            lambda m, p: m.add(app.actions.ActivityDelete, p.selected_products,
                               text="Delete product" if len(p.selected_products) == 1 else "Delete products",
                               enable=len(p.selected_products) > 0 and not
                               database_is_locked(p.db_name) and not
                               database_is_legacy(p.db_name),
                               ),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.CSNew,
                               functional_units=[{prod: m.get_functional_unit_amount(prod)} for prod in p.selected_products],
                               enable=len(p.selected_products) > 0,
                               text="Create setup"
                               ),
            lambda m, p: m.add(app.actions.ActivitySDFToClipboard, p.selected_products,
                               enable=len(p.selected_products) > 0,
                               ),
        ]

        @staticmethod
        def get_functional_unit_amount(key):
            from activity_browser.bwutils.commontasks import refresh_node
            excs = list(refresh_node(key).upstream(["production"]))
            exc = excs[0] if len(excs) == 1 else {}
            return exc.get("amount", 1.0)

    def __init__(self, parent: DatabaseProductsPane, db_name: str):
        """
        Initializes the ProductView.

        Args:
            parent (DatabaseProductsPane): The parent widget.
            db_name (str): The name of the database.
        """
        super().__init__(parent)
        self.setSortingEnabled(True)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(QtWidgets.QTableView.DragDropMode.DragDrop)
        self.setSelectionBehavior(ui.widgets.ABTreeView.SelectionBehavior.SelectRows)
        self.setSelectionMode(ui.widgets.ABTreeView.SelectionMode.ExtendedSelection)

        self.db_name = db_name
        self.pane = parent

        self.propertyDelegate = delegates.PropertyDelegate(self)
        self.overlay = None

    def setDefaultColumnDelegates(self):
        """
        Sets the default column delegates for the view.
        """
        super().setDefaultColumnDelegates()

        columns = self.model().columns()
        for i, col_name in enumerate(columns):
            if not col_name.startswith("property_"):
                continue
            # Set the delegate for property columns
            self.setItemDelegateForColumn(i, self.propertyDelegate)

    def mouseDoubleClickEvent(self, event) -> None:
        """
        Handles the mouse double click event to open the selected activities.

        Args:
            event: The mouse double click event.
        """
        if self.selected_activities:
            app.actions.ActivityOpen.run(self.selected_activities)

    def keyPressEvent(self, event) -> None:
        """
        Handles key press events. Specifically handles Ctrl+C to copy selected data.

        Args:
            event: The key press event.
        """
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            if event.key() == Qt.Key.Key_C:  # Copy
                self.copy_selection_to_clipboard()
                return
            if event.key() == Qt.Key.Key_V:
                self.copy_from_clipboard()
            if event.key() == Qt.Key.Key_A:  # Select All
                self.selectAll()
                return
            if event.key() == Qt.Key.Key_F:  # Find
                self.pane.search_bar.setFocus()
                return
        if event.key() == Qt.Key.Key_Delete:
            if database_is_locked(self.db_name):
                return
            if self.selected_products:
                app.actions.ActivityDelete.run(self.selected_products)
                return

        super().keyPressEvent(event)

    def copy_selection_to_clipboard(self):
        selection = self.selectedIndexes()
        mime_data = self.model().mimeData(selection)

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def copy_from_clipboard(self):
        if database_is_locked(self.db_name):
            return

        clipboard = QtWidgets.QApplication.clipboard()
        mime_data = clipboard.mimeData()

        if mime_data.hasFormat("application/bw-nodekeylist"):
            keys: list = mime_data.retrievePickleData("application/bw-nodekeylist")
            keys = list(set(keys))

            app.actions.ActivityDuplicateToDB.run(keys, self.db_name)

    def dragEnterEvent(self, event):
        """
        Handles the drag enter event.

        Args:
            event: The drag enter event.
        """
        if event.source() == self:
            return

        if database_is_locked(self.db_name):
            return

        if event.mimeData().hasFormat("application/bw-nodekeylist"):
            keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")

            if any(is_node_biosphere(key) for key in keys):
                return

            self.overlay = widgets.ABDropOverlay(self, text="Drop here to duplicate to this database")
            self.overlay.show()
            event.accept()

    def dragMoveEvent(self, event):
        pass

    def dragLeaveEvent(self, event):
        """
        Handles the drag leave event.

        Args:
            event: The drag leave event.
        """
        if self.overlay:
            self.overlay.deleteLater()

    def dropEvent(self, event):
        """
        Handles the drop event.

        Args:
            event: The drop event.
        """
        logger.debug(f"Dropevent from: {type(event.source()).__name__} to: {self.__class__.__name__}")
        # Reset the palette on drop
        if self.overlay:
            self.overlay.deleteLater()

        keys: list = event.mimeData().retrievePickleData("application/bw-nodekeylist")
        keys = list(set(keys))

        app.actions.ActivityDuplicateToDB.run(keys, self.db_name)

    @property
    def selected_products(self) -> list[tuple]:
        """
        Returns the selected products.

        Returns:
            list[tuple]: The list of selected products.
        """
        keys = self.model().values_from_indices("key", self.selectedIndexes())
        types = self.model().values_from_indices("type", self.selectedIndexes())

        return list({key for key, type in zip(keys, types) if not type == "nonfunctional"})

    @property
    def selected_activities(self) -> list[tuple]:
        """
        Returns the selected activities.

        Returns:
            list[tuple]: The list of selected activities.
        """
        processors = self.model().values_from_indices("processor", self.selectedIndexes())
        keys = self.model().values_from_indices("key", self.selectedIndexes())

        return list({processor if not pd.isna(processor) else key for processor, key in zip(processors, keys)})


class ProductModel(ui.core.ABTreeModel):
    #-- flag overrides ---
    def indexDragEnabled(self, index: QtCore.QModelIndex) -> bool:
        return True

    # -- data overrides ---
    def displayData(self, index: QModelIndex) -> any:
        column_name = self.column_name(index)
        if column_name != "node":
            return super().displayData(index)

        row = self.row(index)

        if row is None:
            return None

        # Get the product or name for title
        title = row.get("product") or row.get("name")

        # Build subtitle with name (if product exists) or type
        subtitle_parts = []
        if row.get("product") and row.get("name"):
            # If there's both product and name, show name as subtitle
            subtitle_parts.append(row.get("name"))
        elif row.get("type"):
            # Otherwise show type
            subtitle_parts.append(row.get("type").capitalize())

        subtitle = " | ".join(subtitle_parts) if subtitle_parts else None

        # Build categories list from unit, location, database
        categories = []
        if row.get("unit"):
            categories.append(str(row.get("unit")))
        if row.get("location"):
            categories.append(str(row.get("location")))
        if row.get("key") and isinstance(row.get("key"), tuple):
            categories.append(str(row.get("key")[0]))  # database name

        # Add actual categories if they exist
        node_categories = row.get("categories")
        if node_categories and isinstance(node_categories, (list, tuple)):
            categories.extend([str(cat) for cat in node_categories if str(cat).strip()])

        return {
            "title": title,
            "subtitle": subtitle,
            "categories": categories if categories else None,
        }

    def decorationData(self, index: QtCore.QModelIndex) -> any:
        column_name = self.column_name(index)
        node_type = self.get(index, "type")
        
        if column_name not in ["name", "product", "node"]:
            return None

        if column_name == "name" and node_type in ["product", "waste"]:
            return icons.qicons.process
        if column_name in ["name", "node"] and node_type == "processwithreferenceproduct":
            return icons.qicons.processproduct
        if column_name in ["name", "node"] and node_type in NODETYPES["biosphere"]:
            return icons.qicons.biosphere
        if column_name == "name":
            return icons.qicons.empty

        if column_name in ["product", "node"] and node_type in ["product", "processwithreferenceproduct"]:
            return icons.qicons.product
        if column_name in ["product", "node"] and node_type == "waste":
            return icons.qicons.waste
        return icons.qicons.empty
    
    def toolTipData(self, index: QtCore.QModelIndex) -> str:
        column_name = self.column_name(index)
        if column_name not in ["name", "product"]:
            return None
        
        row = self.row(index)

        html_tooltip = f"""
        <b>{row.get('product')}</b><br>
        <i>{row.get('name')}</i><br>
        <br>
        {row.get('unit')} | {row.get('location')} | {row.get('type')}
        """

        return html_tooltip

    def mimeData(self, indices: list[QtCore.QModelIndex]):
        """
        Returns the mime data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the mime data for.

        Returns:
            core.ABMimeData: The mime data.
        """
        data = core.ABMimeData()
        keys = set(self.values_from_indices("key", indices))
        keys.update(self.values_from_indices("processor", indices))
        keys = {key for key in keys if isinstance(key, tuple)}
        data.setPickleData("application/bw-nodekeylist", list(keys))

        # Add HTML data for Excel with bold formatting
        thread = threading.Thread(target=self.set_excel_nodes_threaded, args=(data, keys))
        thread.start()

        return data

    @staticmethod
    def set_excel_nodes_threaded(data, keys):
        excel_string = nodes_to_excel(list(keys))
        try:
            data.setHtml(excel_string)
        except RuntimeError:
            pass



from qtpy import QtWidgets, QtCore
from loguru import logger

import bw2data as bd
import pandas as pd

from activity_browser import app, app
from activity_browser.ui import widgets, core, delegates


class ImpactCategoriesPane(widgets.ABAbstractPane):
    title = "Impact Categories"
    unique = True

    def __init__(self, parent=None):
        super().__init__(parent)
        self.model = ImpactCategoriesModel(parent=self)
        self.view = ImpactCategoriesView()
        self.view.setModel(self.model)

        self.view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        self.view.setDragEnabled(True)
        self.view.setDragDropMode(QtWidgets.QTableView.DragDropMode.DragOnly)
        self.view.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        self.search = widgets.ABLineEdit(self)
        self.search.setMaximumHeight(30)
        self.search.setPlaceholderText("Quick Search")

        self.search.textChangedDebounce.connect(self.view.setAllFilter)

        self.build_layout()
        self.connect_signals()

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.search)
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)

        self.setLayout(layout)

    def connect_signals(self):
        app.signals.meta.methods_changed.connect(self.sync)
        app.signals.database_read_only_changed.connect(self.sync)

    def sync(self):
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")

        df = self.build_df()
        self.model.set_dataframe(df, group=["_method_name"])

    def build_df(self):
        df = pd.DataFrame(bd.methods.values())
        df["_method_name"] = bd.methods.keys()

        df["name"] = df["_method_name"].apply(lambda x: x[-1])

        cols = ["name", "unit", "num_cfs", "_method_name"]

        if df.empty:
            return pd.DataFrame(columns=cols)

        return df[cols]


class ImpactCategoriesView(widgets.ABTreeView):
    defaultColumnDelegates = {
        "groups": delegates.ListDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(app.actions.MethodNew),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(app.actions.MethodOpen, p.selected_impact_categories,
                               text="Open impact category" if len(p.selected_impact_categories) == 1 else "Open impact categories",
                               enable=len(p.selected_impact_categories) > 0
                               ),
            lambda m, p: m.add(app.actions.MethodDelete, p.selected_impact_categories,
                               text="Delete impact category" if len(
                                   p.selected_impact_categories) == 1 else "Delete impact categories",
                               enable=len(p.selected_impact_categories) > 0
                               ),
            lambda m, p: m.add(app.actions.MethodDuplicate, p.selected_impact_categories,
                               text="Duplicate impact category",
                               enable=len(p.selected_impact_categories) == 1
                               ),
            lambda m, p: m.add(app.actions.MethodRename, p.selected_impact_categories,
                               text="Rename impact category",
                               enable=len(p.selected_impact_categories) == 1
                               ),
        ]

    @property
    def selected_impact_categories(self):
        if not self.selectedIndexes():
            return []
        
        indices = [i for i in self.selectedIndexes() if i.column() == 0]
        impact_categories = []

        for index in indices:
            impact_categories.extend(self.model().get_impact_categories(index))

        return list(set(impact_categories))

    def mouseDoubleClickEvent(self, event) -> None:
        if self.selected_impact_categories:
            app.actions.MethodOpen.run(self.selected_impact_categories)


class ImpactCategoriesModel(core.ABTreeModel):
    """
    A model representing the data for the impact categories.
    """

    def indexDragEnabled(self, index: QtCore.QModelIndex) -> bool:
        """Enable drag for all items."""
        return True

    def mimeData(self, indices: list[QtCore.QModelIndex]):
        """
        Returns the mime data for the given indices.

        Args:
            indices (list[QtCore.QModelIndex]): The indices to get the mime data for.

        Returns:
            core.ABMimeData: The mime data.
        """
        data = core.ABMimeData()
        names = []

        for index in indices:
            names += self.get_impact_categories(index)

        data.setPickleData("application/bw-methodnamelist", list(set(names)))
        return data

    def get_impact_categories(self, index: QtCore.QModelIndex):
        """
        Get all impact category method names for the given index.
        
        For leaf nodes (full depth paths), returns the single method name.
        For branch nodes (partial depth paths), returns all child method names.
        
        Args:
            index: The index to get impact categories for.
            
        Returns:
            list: List of method name tuples.
        """
        if not index.isValid():
            return []
        
        node = index.internalPointer()
        
        if not isinstance(node, core.TreeNode):
            return []
        
        # If this is a leaf node, return its method name
        if node.is_leaf:
            row = self.row(index)
            if row is not None:
                return [row["_method_name"]]
            return []
        
        # If this is a branch node, collect all child method names recursively
        ics = []
        for i, child_node in enumerate(node.children):
            if i >= node.loaded_count:
                break  # Only process loaded children
            child_index = self.createIndex(i, 0, child_node)
            ics += self.get_impact_categories(child_index)
        
        return ics




# -*- coding: utf-8 -*-
import importlib.util
from loguru import logger
from qtpy import QtWidgets

from activity_browser.app import settings
from activity_browser.app.pages.settings.base import BaseSettingsChapter
from activity_browser.ui.icons import qicons


class PluginsSettingsChapter(BaseSettingsChapter):
    """Chapter for plugin-related settings."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # List widget to display enabled plugins
        self.plugin_list = QtWidgets.QListWidget()
        self.plugin_list.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        # Input field for adding new plugins
        self.plugin_input = QtWidgets.QLineEdit()
        self.plugin_input.setPlaceholderText("Enter Python package name (e.g., my_plugin)")

        # Buttons
        self.add_button = QtWidgets.QPushButton("Add")
        self.remove_button = QtWidgets.QPushButton("Remove")
        self.remove_button.setEnabled(False)

        self.build_layout()
        self.connect_signals()
        self.reset()

    def connect_signals(self):
        """Connect signals and slots."""
        self.add_button.clicked.connect(self.add_plugin)
        self.remove_button.clicked.connect(self.remove_plugin)
        self.plugin_input.returnPressed.connect(self.add_plugin)
        self.plugin_list.itemSelectionChanged.connect(self.on_selection_changed)
        self.plugin_list.model().rowsInserted.connect(lambda: self.changed.emit())
        self.plugin_list.model().rowsRemoved.connect(lambda: self.changed.emit())

    def build_layout(self):
        """Build the chapter layout."""
        layout = QtWidgets.QVBoxLayout()

        # Plugin list section
        plugin_group = QtWidgets.QGroupBox("Enabled Plugins")
        plugin_layout = QtWidgets.QVBoxLayout()

        # Description label
        description = QtWidgets.QLabel(
            "Add Python packages that should be imported as plugins.\n"
            "These packages will be loaded when Activity Browser starts."
        )
        description.setWordWrap(True)
        plugin_layout.addWidget(description)

        # List widget
        plugin_layout.addWidget(self.plugin_list)

        # Input section
        input_layout = QtWidgets.QHBoxLayout()
        input_layout.addWidget(self.plugin_input)
        input_layout.addWidget(self.add_button)
        plugin_layout.addLayout(input_layout)

        # Remove button
        plugin_layout.addWidget(self.remove_button)

        plugin_group.setLayout(plugin_layout)

        layout.addWidget(plugin_group)
        layout.addStretch()

        self.setLayout(layout)

    def on_selection_changed(self):
        """Enable/disable remove button based on selection."""
        self.remove_button.setEnabled(len(self.plugin_list.selectedItems()) > 0)

    def module_exists(self, module_name):
        """Check if a module can be found/imported."""
        try:
            spec = importlib.util.find_spec(module_name)
            return spec is not None
        except (ImportError, ModuleNotFoundError, ValueError, AttributeError):
            return False

    def add_plugin_to_list(self, plugin_name):
        """Add a plugin to the list widget with appropriate icon."""
        item = QtWidgets.QListWidgetItem(plugin_name)

        # Check if module exists and add warning icon if not
        if not self.module_exists(plugin_name):
            # Use standard warning icon
            icon = qicons.critical
            item.setIcon(icon)
            item.setToolTip(f"Warning: Module '{plugin_name}' not found. "
                          "Make sure it is installed before starting Activity Browser.")
            logger.warning(f"Plugin module '{plugin_name}' not found")
        else:
            icon = qicons.empty
            item.setIcon(icon)
            item.setToolTip(f"Module '{plugin_name}' is available")

        self.plugin_list.addItem(item)

    def add_plugin(self):
        """Add a plugin to the list."""
        plugin_name = self.plugin_input.text().strip()
        if not plugin_name:
            return

        # Check if plugin already exists
        existing_items = [self.plugin_list.item(i).text() for i in range(self.plugin_list.count())]
        if plugin_name in existing_items:
            QtWidgets.QMessageBox.warning(
                self,
                "Duplicate Plugin",
                f"The plugin '{plugin_name}' is already in the list."
            )
            return

        # Add to list with icon
        self.add_plugin_to_list(plugin_name)
        self.plugin_input.clear()
        logger.debug(f"Added plugin: {plugin_name}")
        self.changed.emit()

    def remove_plugin(self):
        """Remove selected plugin from the list."""
        selected_items = self.plugin_list.selectedItems()
        if not selected_items:
            return

        for item in selected_items:
            plugin_name = item.text()
            row = self.plugin_list.row(item)
            self.plugin_list.takeItem(row)
            logger.debug(f"Removed plugin: {plugin_name}")

        self.changed.emit()

    # --- Settings management methods --- #
    def reset(self):
        """(Re)set to initial values."""
        self.plugin_list.clear()
        enabled_plugins = settings["plugins"].get("enabled_plugins", [])
        for plugin in enabled_plugins:
            self.add_plugin_to_list(plugin)
        self.plugin_input.clear()
        self.remove_button.setEnabled(False)

    def has_changes(self):
        """Check if there are unsaved changes."""
        current_plugins = [self.plugin_list.item(i).text() for i in range(self.plugin_list.count())]
        saved_plugins = settings["plugins"].get("enabled_plugins", [])
        return current_plugins != saved_plugins

    def set_settings(self):
        """Save plugin settings."""
        current_plugins = [self.plugin_list.item(i).text() for i in range(self.plugin_list.count())]
        settings["plugins"]["enabled_plugins"] = current_plugins
        logger.info(f"Saved enabled plugins: {current_plugins}")




# -*- coding: utf-8 -*-
from loguru import logger
from pathlib import Path

from qtpy import QtWidgets

from bw2data import projects

from activity_browser.app import settings, signals

from .startup import StartupSettingsChapter
from .appearance import AppearanceSettingsChapter
from .project_manager import ProjectManagerSettingsChapter
from .metadatastore import MetadataStoreSettingsChapter
from .plugins import PluginsSettingsChapter


class SettingsPage(QtWidgets.QWidget):
    """Settings page with a sidebar navigation for different settings chapters."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("SettingsPage")
        
        # Store initial state for cancel functionality
        self.last_project = projects.current
        self.last_bwdir = projects._base_data_dir
        
        # Chapter list (sidebar)
        self.chapter_list = QtWidgets.QListWidget()
        self.chapter_list.setMaximumWidth(200)
        self.chapter_list.setMinimumWidth(100)
        self.chapter_list.setSpacing(2)
        
        # Stacked widget for chapter content
        self.content_stack = QtWidgets.QStackedWidget()
        
        # Create chapters
        self.startup_chapter = StartupSettingsChapter(self)
        self.appearance_chapter = AppearanceSettingsChapter(self)
        self.project_manager_chapter = ProjectManagerSettingsChapter(self)
        self.metadatastore_chapter = MetadataStoreSettingsChapter(self)
        self.plugins_chapter = PluginsSettingsChapter(self)

        # Add chapters to the stack
        self.chapters = [
            ("Startup", self.startup_chapter),
            ("Appearance", self.appearance_chapter),
            ("Projects", self.project_manager_chapter),
            ("Metadata Store", self.metadatastore_chapter),
            ("Plugins", self.plugins_chapter),
        ]
        
        for name, widget in self.chapters:
            self.chapter_list.addItem(name)
            self.content_stack.addWidget(widget)
        
        # Select first chapter by default
        self.chapter_list.setCurrentRow(0)
        
        # Buttons
        self.button_layout = QtWidgets.QHBoxLayout()
        self.save_button = QtWidgets.QPushButton("Save")
        self.cancel_button = QtWidgets.QPushButton("Cancel")
        self.restore_defaults_button = QtWidgets.QPushButton("Restore Defaults")
        
        self.button_layout.addWidget(self.restore_defaults_button)
        self.button_layout.addStretch()
        self.button_layout.addWidget(self.cancel_button)
        self.button_layout.addWidget(self.save_button)
        
        # Build layout
        self.build_layout()
        self.connect_signals()
        
        # Store initial state and disable save button initially
        self.save_button.setEnabled(False)
    
    def build_layout(self):
        """Build the main layout with sidebar and content area."""
        # Main content area with sidebar and content
        content_layout = QtWidgets.QHBoxLayout()
        content_layout.addWidget(self.chapter_list)
        
        # Add vertical separator
        separator = QtWidgets.QFrame()
        separator.setFrameShape(QtWidgets.QFrame.VLine)
        separator.setFrameShadow(QtWidgets.QFrame.Sunken)
        content_layout.addWidget(separator)
        
        content_layout.addWidget(self.content_stack, 1)
        
        # Main layout
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(5, 5, 5, 5)
        main_layout.addLayout(content_layout, 1)
        main_layout.addLayout(self.button_layout)
        
        self.setLayout(main_layout)
        
        # Set minimum size for resizability
        self.setMinimumSize(400, 300)
    
    def connect_signals(self):
        """Connect signals and slots."""
        signals.project.changed.connect(self.reset_all)

        self.chapter_list.currentRowChanged.connect(self.content_stack.setCurrentIndex)
        self.save_button.clicked.connect(self.save_settings)
        self.cancel_button.clicked.connect(self.cancel_settings)
        self.restore_defaults_button.clicked.connect(self.restore_defaults)
        
        # Connect change signals from each chapter
        for name, chapter in self.chapters:
            if hasattr(chapter, 'changed'):
                chapter.changed.connect(self.on_chapter_changed)

    def on_chapter_changed(self):
        """Called when any chapter's settings change."""
        has_changes = self.has_changes()
        self.save_button.setEnabled(has_changes)
    
    def has_changes(self):
        """Check if any chapter has unsaved changes."""
        for name, chapter in self.chapters:
            if hasattr(chapter, 'has_changes') and chapter.has_changes():
                return True
        return False
    
    def save_settings(self):
        """Save all settings from all chapters."""
        for name, chapter in self.chapters:
            if hasattr(chapter, 'set_settings'):
                chapter.set_settings()
        
        settings.save()
        logger.info("Settings saved successfully")
        
        # Reset all chapters to the new saved state
        self.reset_all()
    
    def cancel_settings(self):
        """Cancel changes and revert to previous state."""
        logger.info("Cancelling settings changes")
        self.reset_all()
            
    def restore_defaults(self):
        """Restore default settings for the current chapter."""
        logger.info("Restoring default settings")
        settings.restore_defaults()
        self.reset_all()
    
    def reset_all(self):
        """Reset all chapters to their initial states."""
        for name, chapter in self.chapters:
            if hasattr(chapter, 'reset'):
                chapter.reset()
        self.save_button.setEnabled(False)




from qtpy import QtWidgets
from qtpy.QtCore import QTimer, Signal, SignalInstance, QStringListModel, Qt
from qtpy.QtGui import QSyntaxHighlighter, QTextCharFormat, QTextDocument, QFont
from qtpy.QtWidgets import QCompleter, QStyledItemDelegate, QStyle


class UnknownWordHighlighter(QSyntaxHighlighter):
    def __init__(self, parent: QTextDocument, known_words: set):
        super().__init__(parent)
        self.known_words = known_words

        # define the format for unknown words
        self.unknown_format = QTextCharFormat()
        self.unknown_format.setUnderlineStyle(QTextCharFormat.SpellCheckUnderline)
        self.unknown_format.setUnderlineColor(Qt.red)

    def highlightBlock(self, text: str):
        if text.startswith("="):
            return
        words = text.split()
        index = 0
        for word in words:
            word_len = len(word)
            if word and word not in self.known_words:
                self.setFormat(index, word_len, self.unknown_format)
            index += word_len + 1  # +1 for the space


class AutoCompleteDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.current_word_index = -1

    def paint(self, painter, option, index):
        text = index.data(Qt.DisplayRole)

        painter.save()

        # Draw selection background if selected
        if option.state & QStyle.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())
            painter.setPen(option.palette.highlightedText().color())
        else:
            painter.setPen(option.palette.text().color())

        # Split text into words and draw each with appropriate font
        words = text.split(" ")
        x = option.rect.x()
        y = option.rect.y()
        spacing = 4  # space between words
        font = option.font
        metrics = painter.fontMetrics()

        for i, word in enumerate(words):
            word_font = QFont(font)
            if i+1 == self.current_word_index:
                word_font.setBold(True)
            painter.setFont(word_font)

            word_width = metrics.horizontalAdvance(word)
            painter.drawText(x, y + metrics.ascent() + (option.rect.height() - metrics.height()) // 2, word)
            x += word_width + spacing
        painter.restore()


class ABTextEdit(QtWidgets.QTextEdit):
    textChangedDebounce: SignalInstance = Signal(str)
    _debounce_ms = 250

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._debounce_timer = QTimer(self, singleShot=True)

        self.textChanged.connect(self._set_debounce)
        self._debounce_timer.timeout.connect(self._emit_debounce)

    def _set_debounce(self):
        self._debounce_timer.setInterval(self._debounce_ms)
        self._debounce_timer.start()

    def _emit_debounce(self):
        self.textChangedDebounce.emit(self.toPlainText())

    def debounce(self):
        return self._debounce_ms

    def setDebounce(self, ms: int):
        self._debounce_ms = ms


class ABAutoCompleTextEdit(ABTextEdit):
    def __init__(self, parent=None, highlight_unknown=False):
        from activity_browser.bwutils.metadata import MetaDataStore  # avoid circular import, should we refactor?

        self.mds = MetaDataStore()
        super().__init__(parent=parent)

        self.auto_complete_word = ""

        # autocompleter settings
        self.model = QStringListModel()
        self.completer = QCompleter(self.model)
        self.completer.setWidget(self)
        self.popup = self.completer.popup()
        self.delegate = AutoCompleteDelegate(self.popup) # set custom delegate to bold the current word
        self.popup.setItemDelegate(self.delegate)
        self.popup.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.completer.setPopup(self.popup)
        self.completer.setCompletionMode(QCompleter.UnfilteredPopupCompletion) # allow all items in popup list
        self.completer.activated.connect(self._insert_auto_complete)

        self.textChanged.connect(self._sanitize_input)
        if highlight_unknown:
            self.highlighter = UnknownWordHighlighter(self.document(), set())
        self.cursorPositionChanged.connect(self._set_autocomplete_items)

    def keyPressEvent(self, event):
        key = event.key()

        if key in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Tab):
            # insert an autocomplete item
            # capture enter/return/tab key
            index = self.popup.currentIndex()
            completion_text = index.data(Qt.DisplayRole)
            self.completer.activated.emit(completion_text)
            return
        elif key in (Qt.Key_Space,):
            self.popup.close()

        super().keyPressEvent(event)

        # trigger on text input keys
        if event.text() or key in (Qt.LeftArrow, Qt.RightArrow):  # filters out non-text keys except l/r arrows
            self._set_autocomplete_items()

    def _sanitize_input(self):
        raise NotImplementedError

    def _set_autocomplete_items(self):
        raise NotImplementedError

    def _insert_auto_complete(self, completion):
        cursor = self.textCursor()
        position = cursor.position()
        completion = completion + " "  # add space to end of new text

        # find where to put cursor back
        new_position = position
        while new_position < len(completion) and completion[new_position] != " ":
            new_position += 1
        new_position += 1  # add one char for space

        # set new text from completion
        self.blockSignals(True)
        self.clear()
        self.setText(completion)
        # set the cursor location
        cursor.setPosition(min(new_position, len(completion)))
        self.setTextCursor(cursor)
        self.blockSignals(False)

        # house keeping
        self._emit_debounce()
        self.popup.close()
        self.auto_complete_word = ""
        self.model.setStringList([])


class MetaDataAutoCompleteTextEdit(ABAutoCompleTextEdit):
    """TextEdit with MetaDataStore completer attached."""
    def __init__(self, parent=None):
        super().__init__(parent=parent, highlight_unknown=True)
        self.database_name = ""

    def _sanitize_input(self):
        if not self.mds.searcher:
            return

        self._debounce_timer.stop()
        text = self.toPlainText()
        clean_text = self.mds.searcher.ONE_SPACE_PATTERN.sub(" ", text)

        if clean_text != text:
            cursor = self.textCursor()
            position = cursor.position()
            self.blockSignals(True)
            self.clear()
            self.insertPlainText(clean_text)
            self.blockSignals(False)
            cursor.setPosition(min(position, len(clean_text)))
            self.setTextCursor(cursor)

        known_words = set()
        for identifier in self.mds.searcher.database_id_manager(self.database_name):
            known_words.update(self.mds.searcher.identifier_to_word[identifier].keys())
        self.highlighter.known_words = known_words

        if len(text) == 0:
            self.popup.close()
        self._set_debounce()

    def _set_autocomplete_items(self):
        if not self.mds.searcher:
            return

        text = self.toPlainText()
        if text.startswith("="):
            self.model.setStringList([])
            self.auto_complete_word = ""
            self.popup.close()
            return

        # find the start and end of the word under the cursor
        cursor = self.textCursor()
        position = cursor.position()
        start = position
        while start > 0 and text[start - 1] != " ":
            start -= 1
        end = position
        while end < len(text) and text[end] != " ":
            end += 1
        current_word = text[start:end]
        if not current_word:
            self.model.setStringList([])
            self.popup.close()
            self.auto_complete_word = ""
            return
        if self.auto_complete_word == current_word:
            # avoid unnecessary auto_complete calls if the current word didnt change
            return
        self.auto_complete_word = current_word

        context = set((text[:start] + text[end:]).split(" "))
        self.delegate.current_word_index = len(text[:start].split(" "))  # current word index for bolding
        # get suggestions for the current word
        suggestions = self.mds.searcher.auto_complete(current_word, context=context, database=self.database_name)
        suggestions = suggestions[:6]  # at most 6, though we should get ~3 usually
        # replace the current word with each alternative
        items = []
        for alt in suggestions:
            new_text = text[:start] + alt + text[end:]
            items.append(new_text)
        if len(items) == 0:
            self.popup.close()
            return

        self.model.setStringList(items)
        # set correct height now that we have data
        max_height = max(
            20,
            self.popup.sizeHintForRow(0) * 3 + 2 * self.popup.frameWidth()
                         )
        self.popup.setMaximumHeight(max_height)
        self.completer.complete()



import os
from pathlib import Path
from loguru import logger

from qtpy import QtGui, QtWidgets, QtCore, PYSIDE6
from qtpy.QtCore import Qt
from qtpy.QtGui import QFontDatabase

from activity_browser.static import fonts, icons


class ABApplication(QtWidgets.QApplication):
    _main_window = None
    _instance = None

    windows = []
        
    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self, *args, **kwargs):
        if self._initialized:
            return

        QtCore.QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)
        QtCore.QCoreApplication.setAttribute(Qt.AA_UseSoftwareOpenGL)

        super().__init__(*args, **kwargs)
        self.set_icon()  # needs to be called right after super().__init__

        QtGui.QGuiApplication.setAttribute(Qt.AA_DontShowIconsInMenus, True)

        self.add_fonts()

        if PYSIDE6:
            self.pyside6_setup()
        
        self._initialized = True

    def add_fonts(self):
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/mono.ttf")
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/ptsans.ttf")
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/notosans.ttf")

    def set_icon(self):
        app_icon = QtGui.QIcon(str(Path(icons.__path__[0]).joinpath("main", "ab-small.png")))
        self.setWindowIcon(app_icon)

    def pyside6_setup(self):
        from qtpy.QtWebEngineQuick import QtWebEngineQuick
        QtWebEngineQuick.initialize()

        style = QtWidgets.QStyleFactory().create("fusion")
        self.setStyle(style)

        self.styleHints().colorSchemeChanged.connect(self.check_palette)
        self.check_palette(self.styleHints().colorScheme())

    def check_palette(self, color_scheme):
        import matplotlib.pyplot as plt

        if color_scheme == Qt.ColorScheme.Dark:
            palette = self.style().standardPalette()
            palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))

            plt.style.use("dark_background")

            os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--force-dark-mode"
        else:
            palette = self.style().standardPalette()

            plt.style.use("default")

            os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = ""
        self.setPalette(palette)

    @property
    def main_window(self) -> QtWidgets.QMainWindow:
        """Returns the main_window widget of the Activity Browser"""
        if self._main_window:
            return self._main_window
        raise Exception(
            "main_window not yet initialized, did you try to access it during startup?"
        )

    @main_window.setter
    def main_window(self, widget: QtWidgets.QMainWindow):
        self._main_window = widget

        # connect global keyboard shortcuts to their respective functions
        for seq, func in _global_shortcuts.items():
            shortcut = QtWidgets.QShortcut(QtGui.QKeySequence(seq), widget)
            shortcut.activated.connect(func)

    def show(self):
        self.main_window.showMaximized()

    def close(self):
        for child in self.children():
            if hasattr(child, "close"):
                child.close()

    def deleteLater(self):
        self.main_window.deleteLater()


def global_shortcut(key_sequence):
    """
    Decorator to register a global keyboard shortcut for the main window. Decorate a function with e.g.
    @global_shortcut("Ctrl+S") to register it as a shortcut. Also works on the run method of actions as long as the
    parameters of said action are taken care of.
    """
    def decorator(func):
        _global_shortcuts[key_sequence] = func
        return func
    return decorator

_global_shortcuts = {}



from itertools import permutations
from collections import Counter, OrderedDict
from logging import getLogger
from time import time
from typing import Optional
import pandas as pd

from activity_browser.bwutils.searchengine import SearchEngine


log = getLogger(__name__)


class MetaDataSearchEngine(SearchEngine):

    # caching for faster operation
    def database_id_manager(self, database):
        if not hasattr(self, "all_database_ids"):
            self.all_database_ids = {}

        if database_ids := self.all_database_ids.get(database):
            self.database_ids = database_ids
            self.current_database = database
        elif database is not None:
            self.database_ids = set(self.df[self.df["database"] == database].index.to_list())
            self.all_database_ids[database] = self.database_ids
            self.current_database = database
        else:
            self.database_ids = None
            self.current_database = "_@@NO_DB_"
        return self.database_ids

    def reset_database_id_manager(self):
        if hasattr(self, "all_database_ids"):
            del self.all_database_ids
        if hasattr(self, "database_ids"):
            del self.database_ids

    def database_word_manager(self, database):
        if not hasattr(self, "all_database_words"):
            self.all_database_words = {}

        if database_words := self.all_database_words.get(database):
            self.database_words = database_words
        elif database is not None:
            ids = self.database_id_manager(database)
            self.database_words = self.reverse_dict_many_to_one({_id: self.identifier_to_word[_id] for _id in ids})
            self.all_database_words[database] = self.database_words
        else:
            self.database_words = None
        return self.database_words

    def reset_database_word_manager(self, database):
        if hasattr(self, "all_database_words") and self.all_database_words.get(database):
            del self.all_database_words[database]
        if hasattr(self, "database_words"):
            del self.database_words

    def database_search_cache(self, database, query, result = None):
        if not hasattr(self, "search_cache"):
            self.search_cache = {}

        if result:
            if self.search_cache.get(database):
                self.search_cache[database][query] = result
            else:
                self.search_cache[database] = {query: result}
            return
        if db_cache := self.search_cache.get(database):
            if cached_result := db_cache.get(query):
                return cached_result
        return

    def reset_search_cache(self, database):
        if hasattr(self, "search_cache") and self.search_cache.get(database):
            del self.search_cache[database]

    def reset_all_caches(self, databases):
        self.reset_database_id_manager()
        for database in databases:
            self.reset_database_word_manager(database)
            self.reset_search_cache(database)

    def add_identifier(self, data: pd.DataFrame) -> None:
        super().add_identifier(data)
        self.reset_all_caches(data["database"].unique())

    def remove_identifiers(self, identifiers, logging=True) -> None:
        t = time()

        identifiers = set(identifiers)
        current_identifiers = set(self.df.index.to_list())
        identifiers = identifiers | current_identifiers  # only remove identifiers currently in the data
        databases = self.df.loc[identifiers, ["databases"]].unique()  # extract databases for cache cleaning
        if len(identifiers) == 0:
            return

        for identifier in identifiers:
            super().remove_identifier(identifier, logging=False)

        if logging:
            log.debug(f"Search index updated in {time() - t:.2f} seconds "
                      f"for {len(identifiers)} removed items ({len(self.df)} items ({self.size_of_index()}) currently).")
        self.reset_all_caches(databases)

    def change_identifier(self, identifier, data: pd.DataFrame) -> None:
        super().change_identifier(identifier, data)
        self.reset_all_caches(data["database"].unique())

    def auto_complete(self, word: str, context: Optional[set] = set(), database: Optional[str] = None) -> list:
        """Based on spellchecker, make more useful for autocompletions
        """
        def word_to_identifier_to_word(check_word):
            if len(context) == 0:
                return 1
            multiplier = 1
            for identifier in self.word_to_identifier[check_word]:
                for context_word in context:
                    for spell_checked_context_word in spell_checked_context[context_word]:
                        if spell_checked_context_word in self.identifier_to_word[identifier]:
                            multiplier += 1
                    if context_word not in self.word_to_identifier.keys():
                        continue
                    if context_word in self.identifier_to_word[identifier]:
                        multiplier += 4
            return multiplier

        # count occurrences of a word, count double so word_to_identifier_to_word will never multiply by 1
        count_occurrence = lambda x: sum(self.word_to_identifier[x].values()) * 2

        if len(word) <= 1:
            return []

        self.database_id_manager(database)

        if len(context) > 0:
            spell_checked_context = {}
            for context_word in context:
                spell_checked_context[context_word] = self.spell_check(context_word).get(context_word, [])[:5]

        matches_min = 2  # ideally we have at least this many alternatives
        matches_max = 4  # ideally don't much more than this many matches
        never_accept_this = 4  # values this edit distance or over always rejected
        # or max 2/3 of len(word) if less than never_accept_this
        never_accept_this = int(round(max(1, min((len(word) * 0.66), never_accept_this)), 0))

        # first, find possible matches quickly
        q_grams = self.text_to_positional_q_gram(word)
        possible_matches = self.find_q_gram_matches(set(q_grams), return_all=True)

        first_matches = Counter()
        other_matches = {}
        probably_keys = Counter()  # if we suspect it's a key hash, dump it at the end of the list

        # now, refine with edit distance
        for row in possible_matches.itertuples():
            if word == row[1]:
                continue
            # find edit distance of same size strings
            edit_distance = self.osa_distance(word, row[1][:len(word)], cutoff=never_accept_this)
            if len(row[1]) == 32 and edit_distance <= 1:
                probably_keys[row[1]] = 100 - edit_distance  # keys need to be sorted on edit distance, not on occurence
            elif edit_distance == 0:
                first_matches[row[1]] = count_occurrence(row[1]) * word_to_identifier_to_word(row[1])
            elif edit_distance < never_accept_this and len(first_matches) < matches_min:
                if not other_matches.get(edit_distance):
                    other_matches[edit_distance] = Counter()
                other_matches[edit_distance][row[1]] = count_occurrence(row[1]) * word_to_identifier_to_word(row[1])
            else:
                continue

        # add matches in correct order:
        matches = [match for match, _ in first_matches.most_common()]
        # if we have fewer matches than goal, add more 'less good' matches
        if len(matches) < matches_min:
            for i in range(1, never_accept_this):
                # iteratively increase matches with 'worse' results so we hit goal of minimum alternatives
                if new := other_matches.get(i):
                    prev_num = 10e100
                    for match, num in new.most_common():
                        if num == prev_num:
                            matches.append(match)
                        elif num != prev_num and len(matches) <= matches_max:
                            matches.append(match)
                        else:
                            break
                        prev_num = num

        matches = matches + [match for match, _ in probably_keys.most_common()]
        return matches

    def find_q_gram_matches(self, q_grams: set, return_all: bool = False) -> pd.DataFrame:
        """Overwritten for extra database specific reduction of results.
        """
        n_q_grams = len(q_grams)

        matches = {}

        # find words that match our q-grams
        for q_gram in q_grams:
            if words := self.q_gram_to_word.get(q_gram, False):
                # q_gram exists in our search index
                for word in words:
                    if isinstance(self.database_ids, set):
                        # DATABASE SPECIFIC now filter on whether word is in the database
                        in_db = False
                        for _id in self.word_to_identifier[word]:
                            if _id in self.database_ids:
                                in_db = True
                                break
                    else:
                        in_db = True
                    if in_db:
                        matches[word] = matches.get(word, 0) + words[word]

        # if we find no results, return an empty dataframe
        if len(matches) == 0:
            return pd.DataFrame({"word": [], "matches": []})

        # otherwise, create a dataframe and
        # reduce search results to most relevant results
        matches = {"word": matches.keys(), "matches": matches.values()}
        matches = pd.DataFrame(matches)
        max_q = max(matches["matches"])  # this has the most matching q-grams

        # determine how many results we want to keep based on how good our results are
        if not return_all:
            min_q = min(max(max_q * 0.32,  # have at least a third of q-grams of best match or...
                            max(n_q_grams * 0.5,  # if more, at least half the q-grams in the query word?
                                1)),  # okay just do 1 q-gram if there are no more in the word
                        max_q)  # never have min_q be over max_q
        else:
            min_q = 0

        matches = matches[matches["matches"] >= min_q]
        matches = matches.sort_values(by="matches", ascending=False)
        matches = matches.reset_index(drop=True)

        return matches.iloc[:min(len(matches), 2500), :]  # return at most this many results

    def search_size_1(self, queries: list, original_words: set, orig_word_weight=5, exact_word_weight=1) -> dict:
        """Return a dict of {query_word: Counter(identifier)}.

        queries: is a list of len 1 tuple/lists of words that are a searched word or a 'spell checked' similar word
        original words: a list of words actually searched for (not including spellchecked)

        orig_word_weight: additional weight to add to original words
        exact_word_weight: additional weight to add to exact word matches (as opposed to be 'in' str)

        First, we find all matching words, creating a dict of words in 'queries' as keys and words matching that query word as list of values
        Next, we convert this to identifiers and add weights:
            Weight will be increased if matching 'orig_word_weight' or 'exact_word_weight'
        """
        matches = {}
        t2 = time()
        # add each word in search index if query_word in word
        for word in self.database_words.keys():
            for query in queries:
                # query is list/tuple of len 1
                query_word = query[0]  # only use the word
                if query_word in word:
                    words = matches.get(query_word, [])
                    words.extend([word])
                    matches[query_word] = words

        # now convert matched words to matched identifiers
        matched_identifiers = {}
        for word, matching_words in matches.items():
            if result := self.database_search_cache(self.current_database, word):
                matched_identifiers[word] = result
                continue
            id_counter = matched_identifiers.get(word, Counter())
            for matched_word in matching_words:
                weight = self.base_weight

                # add the word n times, where n is the weight, original search word is weighted higher than alternatives
                if matched_word in original_words:
                    weight += orig_word_weight  # increase weight for original word
                if matched_word == word:
                    weight += exact_word_weight  # increase weight for exact matching word

                id_counter = self.weigh_identifiers(self.database_words[matched_word], weight, id_counter)
                matched_identifiers[word] = id_counter
            self.database_search_cache(self.current_database, word, matched_identifiers[word])

        return matched_identifiers

    def fuzzy_search(self, text: str, database: Optional[str] = None, return_counter: bool = False, logging: bool = True) -> list:
        """Overwritten for extra database specific reduction of results.
        """
        t = time()
        text = text.strip()

        if len(text) == 0:
            log.debug(f"Empty search, returned all items")
            return self.df.index.to_list()

        # DATABASE SPECIFIC get the set of ids that is in this database
        self.database_id_manager(database)
        self.database_word_manager(database)

        queries = self.build_queries(text)

        # make list of unique original words
        orig_words = OrderedDict()
        for word in text.split(" "):
            orig_words[word] = False
        orig_words = orig_words.keys()
        orig_words = {self.clean_text(word) for word in orig_words}

        # order the queries by the amount of words they contain
        # we do this because longer queries (more words) are harder to find, but we have many alternatives so we search in a smaller search space
        queries_by_size = OrderedDict()
        longest_query = max([len(q) for q in queries])
        for query_len in range(1, longest_query + 1):
            queries_by_size[query_len] = [q for q in queries if len(q) == query_len]

        # first handle queries of length 1
        query_to_identifier = self.search_size_1(queries_by_size[1], orig_words)

        # DATABASE SPECIFIC ensure all identifiers are in the database
        if isinstance(self.database_ids, set):
            new_q2i = {}
            for word, _ids in query_to_identifier.items():
                keep = set.intersection(set(_ids.keys()), self.database_ids)
                new_id_counter = Counter()
                for _id in keep:
                    new_id_counter[_id] = _ids[_id]
                if len(new_id_counter) > 0:
                    new_q2i[word] = new_id_counter
            query_to_identifier = new_q2i

        # get all results into a df, we rank further later
        all_identifiers = set()
        for id_list in [id_list for id_list in query_to_identifier.values()]:
            all_identifiers.update(id_list)
        search_df = self.df.loc[list(all_identifiers)]

        # now, we search for combinations of query words and get only those identifiers
        # we then reduce de search_df further for only those matching identifiers
        # we then search the permutations of that set of words
        for q_len, query_set in queries_by_size.items():
            if q_len == 1:
                # we already did these above
                continue
            for query in query_set:
                # get the intersection of all identifiers
                # meaning, a set of identifiers that occur in ALL sets of len(1) for the individual words in the query
                # this ensures we only ever search data where ALL items occur to substantially reduce search-space
                # finally, make this a Counter (with each item=1) so we can properly weigh things later
                query_id_sets = [set(query_to_identifier.get(q_word)) for q_word in query if
                                 query_to_identifier.get(q_word, False)]
                if len(query_id_sets) == 0:
                    continue
                query_identifier_set = set.intersection(*query_id_sets)
                if len(query_identifier_set) == 0:
                    # there is no match for this combination of query words, skip
                    break

                # now we convert the query identifiers to a Counter of 'occurrence',
                # where we weigh queries with only original words higher
                query_identifiers = Counter()
                for identifier in query_identifier_set:
                    weight = 0
                    for query_word in query:
                        # if the query_word and identifier combination exist get score, otherwise 0
                        weight += query_to_identifier.get(query_word, {}).get(identifier, 0)

                    query_identifiers[identifier] = weight

                # we now add these identifiers to a counter for this query name,
                query_name = " ".join(query)

                weight = self.base_weight * q_len
                query_to_identifier[query_name] = self.weigh_identifiers(query_identifiers, weight, Counter())

                # now search for all permutations of this query combined with a space
                query_df = search_df[search_df[self.identifier_name].isin(query_identifiers)]
                for query_perm in permutations(query):
                    query_perm_str = " ".join(query_perm)
                    if result := self.database_search_cache(self.current_database, query_perm_str):
                        new_ids = result
                    else:
                        mask = self.filter_dataframe(query_df, query_perm_str, search_columns=["query_col"])
                        new_df = query_df.loc[mask].reset_index(drop=True)
                        if len(new_df) == 0:
                            # there is no match for this permutation of words, skip
                            continue
                        new_id_list = new_df[self.identifier_name]

                        new_ids = Counter()
                        for new_id in new_id_list:
                            new_ids[new_id] = query_identifiers[new_id]
                        self.database_search_cache(self.current_database, query_perm_str, new_ids)
                    # we weigh a combination of words that is next also to each other even higher than just the words separately
                    query_to_identifier[query_name] = self.weigh_identifiers(new_ids, weight,
                                                                             query_to_identifier[query_name])
        # now finally, move to one object sorted list by highest score
        all_identifiers = Counter()
        for identifiers in query_to_identifier.values():
            all_identifiers += identifiers

        if return_counter:
           return_this = all_identifiers
        else:
            # now sort on highest weights and make list type
            return_this = [identifier[0] for identifier in all_identifiers.most_common()]
        if logging:
            log.debug(
                f"Found {len(all_identifiers)} search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
        return return_this

    def search(self, text, database: Optional[str] = None) -> list:
        """Search the dataframe on this text, return a sorted list of identifiers."""
        t = time()
        text = text.strip()

        if len(text) == 0:
            log.debug(f"Empty search, returned all items")
            return self.df.index.to_list()

        # get the set of ids that is in this database
        self.database_id_manager(database)

        fuzzy_identifiers = self.fuzzy_search(text, database=database, logging=False)
        if len(fuzzy_identifiers) == 0:
            log.debug(f"Found 0 search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
            return []

        # take the fuzzy search sub-set of data and search it literally
        df = self.df.loc[fuzzy_identifiers].copy()

        literal_identifiers = self.literal_search(text, df)
        if len(literal_identifiers) == 0:
            log.debug(
                f"Found {len(fuzzy_identifiers)} search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
            return fuzzy_identifiers

        # append any fuzzy identifiers that were not found in the literal search
        literal_id_set = set(literal_identifiers)
        remaining_fuzzy_identifiers = [
            _id for _id in fuzzy_identifiers if _id not in literal_id_set]
        identifiers = literal_identifiers + remaining_fuzzy_identifiers

        log.debug(
            f"Found {len(identifiers)} ({len(literal_identifiers)} literal) search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
        return identifiers



from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt


class ABCloseButton(QtWidgets.QWidget):
    """Custom close button with hover effect."""
    clicked: QtCore.SignalInstance = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)


        self.label = QtWidgets.QLabel("×", self)

        self.label.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Weight.Light))
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setFixedSize(16, 16)
        self.label.mousePressEvent = lambda event: self.clicked.emit()

        self.label.setStyleSheet("""
            QLabel {
                border-radius: 8px;
                background-color: transparent;
            }
            QLabel:hover {
                background-color: rgba(255, 0, 0, 0.5);
            }
        """)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(5, 0, 0, 0)
        layout.addWidget(self.label)
        self.setLayout(layout)


class ABMinimizeButton(QtWidgets.QWidget):
    """Custom close button with hover effect."""
    clicked: QtCore.SignalInstance = QtCore.Signal()

    def __init__(self, parent=None):
        super().__init__(parent)

        self.label = QtWidgets.QLabel("-", self)

        self.label.setFont(QtGui.QFont("Arial", 12, QtGui.QFont.Weight.Light))
        self.label.setAlignment(Qt.AlignCenter)
        self.label.setFixedSize(16, 16)
        self.label.mousePressEvent = lambda event: self.clicked.emit()

        self.setStyleSheet("""
            QLabel {
                border-radius: 8px;
                background-color: transparent;
            }
            QLabel:hover {
                background-color: rgba(42, 157, 244, 0.5);
            }
        """)

        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(5, 0, 0, 0)
        layout.addWidget(self.label)
        self.setLayout(layout)



# -*- coding: utf-8 -*-
from loguru import logger
from qtpy import QtWidgets

from activity_browser.app import settings
from activity_browser.app.pages.settings.base import BaseSettingsChapter


class AppearanceSettingsChapter(BaseSettingsChapter):
    """Chapter for appearance-related settings."""

    theme_map = {
        "default": "System default",
        "light": "Light theme",
        "dark": "Dark theme",
    }
    
    pane_tab_position_map = {
        "top": "Top",
        "bottom": "Bottom",
        "left": "Left",
        "right": "Right",
    }
    
    def __init__(self, parent=None):
        super().__init__(parent)
        
        # Theme selector
        self.theme_combo = QtWidgets.QComboBox()
        
        # Pane tab position selector
        self.pane_tab_position_combo = QtWidgets.QComboBox()
        
        self.build_layout()
        self.connect_signals()
        self.reset()
    
    def connect_signals(self):
        """Connect signals and slots."""
        # Emit changed signal when settings change
        self.theme_combo.currentTextChanged.connect(lambda: self.changed.emit())
        self.pane_tab_position_combo.currentTextChanged.connect(lambda: self.changed.emit())
    
    def build_layout(self):
        """Build the chapter layout."""
        layout = QtWidgets.QVBoxLayout()
        
        # Theme section
        theme_group = QtWidgets.QGroupBox("Theme")
        theme_layout = QtWidgets.QGridLayout()
        theme_layout.addWidget(QtWidgets.QLabel("Theme:"), 0, 0)
        theme_layout.addWidget(self.theme_combo, 0, 1)
        theme_group.setLayout(theme_layout)
        
        # Pane tab position section
        pane_tab_group = QtWidgets.QGroupBox("Pane Tab Position")
        pane_tab_layout = QtWidgets.QGridLayout()
        pane_tab_layout.addWidget(QtWidgets.QLabel("Position:"), 0, 0)
        pane_tab_layout.addWidget(self.pane_tab_position_combo, 0, 1)
        pane_tab_group.setLayout(pane_tab_layout)
        
        layout.addWidget(theme_group)
        layout.addWidget(pane_tab_group)
        layout.addStretch()
        
        self.setLayout(layout)
    
    # --- Settings management methods --- #
    def reset(self):
        """(Re)set to initial values."""
        self.theme_combo.clear()
        self.theme_combo.addItems(self.theme_map.values())
        self.theme_combo.setCurrentText(self.theme_map.get(settings["appearance"]["theme"], "System default"))
        
        self.pane_tab_position_combo.clear()
        self.pane_tab_position_combo.addItems(self.pane_tab_position_map.values())
        self.pane_tab_position_combo.setCurrentText(self.pane_tab_position_map.get(settings["appearance"]["pane_tab_position"], "Bottom"))

    def has_changes(self):
        """Check if there are unsaved changes."""
        current_state = {
            'theme': self.theme_combo.currentText(),
            'pane_tab_position': self.pane_tab_position_combo.currentText(),
        }
        initial_state = {
            'theme': self.theme_map.get(settings["appearance"]["theme"], "System default"),
            'pane_tab_position': self.pane_tab_position_map.get(settings["appearance"]["pane_tab_position"], "Bottom"),
        }
        return current_state != initial_state
    
    def set_settings(self):
        """Save appearance settings."""
        new_theme = self.theme_combo.currentText()
        settings["appearance"]["theme"] = [key for key, value in self.theme_map.items() if value == new_theme][0]
        
        new_pane_position = self.pane_tab_position_combo.currentText()
        settings["appearance"]["pane_tab_position"] = [key for key, value in self.pane_tab_position_map.items() if value == new_pane_position][0]




import itertools
import functools
import re
from collections import Counter, OrderedDict, defaultdict
from typing import Iterable, Optional
from time import time

from loguru import logger

import pandas as pd
import numpy as np


class SearchEngine:
    """
    A Search Engine class, takes a dataframe and makes it searchable.

    A search requires a string, and will return a list of unique identifiers in the dataframe.
    There are three options for search:
        SearchEngine.literal_search(): searches for exact matches of the search query
        SearchEngine.fuzzy_search(): searches for approximate matches of search query, sorted by relevance
        SearchEngine.search(): combines both of the above, literal matches are returned first, next all fuzzy results,
        but subsets sorted by relevance.
    It is recommended to always use searchEngine.search(), but the other options are there.

    Initialization takes:
        df: Dataframe that needs to be searchable.
        identifier_name: values in this column will be returned as search results, all values in this column need to be unique.
        searchable_columns: these columns need to be searchable, if none are given, all columns will be made searchable.

    Updating data is possible as well:
        add_identifier(): adds this identifier to the searchable data
        remove_identifier(): removes this identifier from the searchable data
        change_identifier(): changes this identifier (wrapper for remove_identifier and add_identifier)

    """

    def __init__(self, df: pd.DataFrame, identifier_name: str, searchable_columns: list = []):
        t = time()
        logger.debug(f"SearchEngine initializing for {len(df)} items")

        # compile regex patterns for cleaning
        self.SUB_END_PATTERN = re.compile(r"[,.\"'`)\[\]}\\/\-−_:;+…]+(?=\s|$)")  # remove these from end of word
        self.SUB_START_PATTERN = re.compile(r"(?:^|\s)[,.\"'`(\[{\\/\-−_:;+]+")  # remove these from start of word
        self.ONE_SPACE_PATTERN = re.compile(r"\s+")  # remove these multiple whitespaces

        self.q = 2  # character length of q grams
        self.base_weight = 10  # base weighting for sorting results

        if identifier_name not in df.columns:  # make sure identifier col exist
            raise NameError(f"Identifier column {identifier_name} not found in dataframe. Use an existing column name.")
        if df[identifier_name].nunique() != df.shape[0]:  # make sure identifiers are all unique
            raise KeyError(
                f"Identifier column {identifier_name} must only contain unique values. Found {df[identifier_name].nunique()} unique values for length {df.shape[0]}")

        self.identifier_name = identifier_name

        # ensure columns given actually exist
        # always ensure "identifier" is present
        if searchable_columns == []:
            # if no list is given, assume all columns are searchable
            self.columns = list(df.columns)
        else:
            # create subset of columns to be searchable, discard rest
            self.columns = [col for col in searchable_columns if col in df.columns]
            if self.identifier_name not in self.columns:  # keep identifier col
                self.columns.append(self.identifier_name)
            df = df[self.columns]
        # set the identifier column as index
        df = df.set_index(self.identifier_name, drop=False)

        # convert all data to str
        df = df.astype(str)

        # find the self.identifier_name column index and store as int
        self.identifier_column = self.columns.index(self.identifier_name)

        # store all searchable column indices except the identifier
        self.searchable_columns = [i for i in range(len(self.columns)) if i != self.identifier_column]

        # initialize search index dicts and update df
        self.identifier_to_word = {}
        self.word_to_identifier = {}
        self.word_to_q_grams = {}
        self.q_gram_to_word = {}
        self.df = pd.DataFrame()

        self.update_index(df)

        logger.debug(f"SearchEngine Initialized in {time() - t:.2f} seconds")

    #   +++ Utility functions

    def update_index(self, update_df: pd.DataFrame) -> None:
        """Update search index dicts and the df."""

        def update_dict(update_me: dict, new: dict) -> dict:
            """Update a dict of counters with new dict of counters."""
            # set to empty set if we know update_me is empty, otherwise, find set intersection
            update_keys = set() if len(update_me) == 0 else new.keys() & update_me.keys()
            if len(update_keys) == 0:
                new_data = new
            else:
                for update_key in update_keys:
                    update_me[update_key].update(new[update_key])
                new_data = {key: value for key, value in new.items() if key not in update_keys}
            # finally add any completely new data
            # update_me.update(new_data)
            update_me = update_me | new_data
            return update_me

        if len(update_df) == 0:
            return

        t = time()
        size_old = len(self.df)
        # identifier to word and df
        i2w, update_df = self.words_in_df(update_df)
        self.identifier_to_word = update_dict(self.identifier_to_word, i2w)
        self.df = pd.concat([self.df, update_df])
        # word to identifier
        w2i = self.reverse_dict_many_to_one(i2w)
        self.word_to_identifier = update_dict(self.word_to_identifier, w2i)
        # word to q-gram
        w2q = self.list_to_q_grams(w2i.keys())
        self.word_to_q_grams = update_dict(self.word_to_q_grams, w2q)
        # q-gram to word
        q2w = self.reverse_dict_many_to_one(w2q)
        self.q_gram_to_word = update_dict(self.q_gram_to_word, q2w)
        size_new = len(self.df)
        size_dif = size_new - size_old
        logger.debug(f"Search index updated in {time() - t:.2f} seconds.")

    def clean_text(self, text: str):
        """Clean a string so it doesn't contain weird characters or multiple spaces etc."""
        text = text.lower()
        text = self.SUB_END_PATTERN.sub("", text)
        text = self.SUB_START_PATTERN.sub(" ", text)

        text = self.ONE_SPACE_PATTERN.sub(" ", text).strip()
        return text

    def text_to_positional_q_gram(self, text: str) -> list:
        """Return a positional list of q-grams for the given string.

        q-grams are n-grams on character level.
        q-grams at q=2 of "word" would be "wo", "or" and "rd"
        https://en.wikipedia.org/wiki/N-gram

        Note: these are technically _positional_ q-grams, but we don't use their positions currently.
        """
        q = self.q
        n = len(text)
        # just return a single-item list if the text is equal or shorter than q
        # else, generate q-grams
        if n <= q:
            return [text]
        return list(text[i:i + q] for i in range(n - q + 1))

    def df_clean(self, df):
        """Clean the text in query_col.

        apply multi-processing when the computer is able and its relevant
        """
        df["query_col"] = df["query_col"].apply(self.clean_text)
        return df

    def words_in_df(self, df: pd.DataFrame = None) -> tuple[dict, pd.DataFrame]:
        """Return a dict of {identifier: word} for df."""

        df = df if df is not None else self.df.copy()
        df = df.fillna("")  # avoid nan
        # assemble query_col
        df["query_col"] = df.iloc[:, self.searchable_columns].astype(str).agg(" | ".join, axis=1)
        # clean all text at once using vectorized operations
        df["query_col"] = self.df_clean(df.loc[:, ["query_col"]])
        # build the identifier_word_dict dictionary - filter out empty strings
        identifier_word_dict = df["query_col"].apply(
            lambda text: Counter(word for word in text.split(" ") if word)
        ).to_dict()
        return identifier_word_dict, df

    def reverse_dict_many_to_one(self, dictionary: dict) -> dict:
        """Reverse a dictionary of Counter objects."""
        reverse = defaultdict(Counter)
        for identifier, counter_object in dictionary.items():
            if not isinstance(counter_object, Counter):
                logger.warning(f"Skipping non-Counter object for {identifier}: {type(counter_object)}")
                continue
            for countable, count in counter_object.items():
                if countable:  # skip empty strings
                    reverse[countable][identifier] += count
        return dict(reverse)

    def list_to_q_grams(self, word_list: Iterable) -> dict:
        """Convert a list of unique words to a dict with Counter objects.

        Number will be the occurrences of that q-gram in that word.

        return = {
            "word": Counter(
                "wo": 1
                "or": 1
                "rd": 1
                ),
            ...
            }
        """
        text_to_q_gram = self.text_to_positional_q_gram
        return {
            word: Counter(text_to_q_gram(word))
            for word in word_list
        }

    def word_in_index(self, word: str) -> bool:
        """Convenience function to check if a single word is in the search index."""
        if " " in word:
            raise Exception(
                f"Given word '{word}' must not contain spaces.")
        return word in self.word_to_identifier.keys()

    #   +++ Changes to searchable data

    def add_identifier(self, data: pd.DataFrame) -> None:
        """Add this data to the search index.

        identifier column is REQUIRED to be present
        ALL data in the given dataframe will be added, if columns should not be added, they should be removed before
        calling this function
        """

        # ensure we have identifier column
        if self.identifier_name not in data.columns:
            raise Exception(
                f"Identifier column '{self.identifier_name}' not in new data, impossible to add data without identifier")

        # make sure we the new identifiers do not yet exist
        existing_ids = set(self.df.index.to_list())
        for identifier in data[self.identifier_name]:
            if identifier in existing_ids:
                raise Exception(
                    f"Identifier '{identifier}' is already in use, use a different identifier or use the change_identifier function.")

        # make sure all new identifiers given are unique
        if data[self.identifier_name].nunique() != data.shape[0]:
            raise KeyError(
                f"Identifier column {self.identifier_name} must only contain unique values. Found {data[self.identifier_name].nunique()} unique values for length {data.shape[0]}")

        df_cols = self.columns
        # add cols to new data that are missing
        for col in df_cols:
            if col not in data.columns:
                data.loc[:, col] = [""] * len(data)
        # re-order cols, first existing, then new
        df_col_set = set(df_cols)
        new_cols = [col for col in data.columns if col not in self.columns if col not in df_col_set]
        data_cols = df_cols + new_cols
        data = data[data_cols]  # re-order new data to be in correct order

        # add cols from new data to correct places
        self.columns.extend(new_cols)
        self.searchable_columns.extend([i for i, col in enumerate(data_cols) if col in new_cols])

        # convert df
        data = data.set_index(self.identifier_name, drop=False)
        data = data.astype(object).fillna("")
        data = data.astype(str)

        # update the search index data
        self.update_index(data)

    def remove_identifier(self, identifier, logging=True) -> None:
        """Remove this identifier from self.df and the search index.
        """
        if logging:
            t = time()

        # make sure the identifier exists
        if identifier not in self.df.index.to_list():
            logger.warning(
                f"Identifier '{identifier}' does not exist in the search data, cannot remove identifier that do not exist."
            )
            return

        self.df = self.df.drop(identifier)

        # find words that may need to be removed
        words = self.identifier_to_word[identifier]
        for word in words:
            if len(self.word_to_identifier[word]) == 1:
                # this word is only found in this identifier,
                # remove the word and check for q grams
                del self.word_to_identifier[word]

                q_grams = self.word_to_q_grams[word]
                for q_gram in q_grams:
                    if len(self.q_gram_to_word[q_gram]) == 1:
                        # this q_gram is only used in this word,
                        #  remove it
                        del self.q_gram_to_word[q_gram]
                    elif len(self.q_gram_to_word[q_gram]) > 1:
                        # this q_gram is used in multiple words, only remove the word from the q_gram
                        del self.q_gram_to_word[q_gram][word]

                del self.word_to_q_grams[word]
            else:
                # this word is found in multiple identifiers
                # word_to_q_gram and q_gram_to_word do not need to be changed, the word still exists
                # remove the identifier the word in word_to_identifier
                del self.word_to_identifier[word][identifier]
        # finally, remove the identifier
        del self.identifier_to_word[identifier]

        if logging:
            logger.debug(f"Search index updated in {time() - t:.2f} seconds "
                         f"for 1 removed item ({len(self.df)}.")

    def change_identifier(self, identifier, data: pd.DataFrame) -> None:
        """Change this identifier.

        identifier must be an identifier that is in use
        data must be a dataframe of 1 row with all change data
        data is overwritten with the new data in 'data', columns not given remain unchanged
        """

        # make sure only 1 change item is given
        if len(data) > 1 or len(data) < 1:
            raise Exception(
                f"change data must be for exactly 1 identifier, but {len(data)} items were given.")
        # make sure correct use of identifier
        if identifier not in self.df.index.to_list():
            raise Exception(
                f"Identifier '{identifier}' does not exist in the search data, use an existing identifier or use the add_identifier function.")
        if self.identifier_name in data.columns and data[self.identifier_name].to_list() != [identifier]:
            raise Exception(
                "Identifier field cannot be changed, first remove item and then add new identifier")
        if "query_col" in data.keys():
            logger.debug(
                f"Field 'query_col' is a protected field for search engine and will be ignored for changing {identifier}")


        # overwrite new data where relevant
        update_data = self.df.loc[[identifier], self.columns]
        data = data.reset_index(drop=True)
        for col in data.columns:
            value = data.loc[0, col]
            update_data[col] = [value]

        # remove the entry
        self.remove_identifier(identifier, logging=False)
        # add entry with updated data
        self.add_identifier(update_data)

    #   +++ Search

    def filter_dataframe(self, df: pd.DataFrame, pattern: str, search_columns: Optional[list] = None) -> pd.Series:
        """Filter the search columns of a dataframe on a pattern.

        Returns a mask (true/false) pd.Series with matching items."""

        search_columns = search_columns if search_columns else self.columns
        mask = functools.reduce(
            np.logical_or,
            [
                df[col].apply(lambda x: pattern in x.lower())
                for col in search_columns
            ],
        )
        return mask

    def literal_search(self, text, df: Optional[pd.DataFrame] = None) -> list:
        """Do literal search of the text in all original columns that were given."""

        if df is None:
            df = self.df.copy()

        identifiers = self.filter_dataframe(df, text)
        df = df.loc[identifiers]
        identifiers = df.index.to_list()
        return identifiers

    def osa_distance(self, word1: str, word2: str, cutoff: int = 0, cutoff_return: int = 1000) -> int:
        """Calculate the Optimal String Alignment (OSA) edit distance between two strings, return edit distance.

        Has additional cutoff variable, if cutoff is higher than 0 and if the words have
        a larger edit distance, return a large number (note: cutoff <= edit_dist, not cutoff < edit_dist)

        OSA is a restricted form of the Damerau–Levenshtein distance.
        https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance#Optimal_string_alignment_distance

        The edit distance is how many operations (insert, delete, substitute or transpose a character) need to happen to convert one string to another.
        insert and delete are obvious operations, but substitute and transpose are explained:
            substitute: replace one character with another: e.g. word1='cat' word2='cab', 't'->'b' substitution is 1 operation
            transpose: swap the places of two adjacent characters with each other: e.g. word1='coal' word2='cola' 'al' -> 'la' transposition is 1 operation

        The minimum amount of edit operations (OSA edit distance) is returned.
        """
        if word1 == word2:
            # if the strings are the same, immediately return 0
            return 0

        len1, len2 = len(word1), len(word2)

        if 0 < cutoff <= abs(len1 - len2):
            # if the length difference between 2 words is over the cutoff,
            # just return instead of calculating the edit distance
            return cutoff_return

        if len1 == 0 or len2 == 0:
            # in case (at least) one of the strings is empty,
            # return the length of the longest string
            return max(len1, len2)

        if len1 < len2 and cutoff > 0:
            # make sure word1 is always the longest (required for early stopping with cutoff)
            word1, word2 = word2, word1
            len1, len2 = len2, len1

        # Initialize matrix
        distance = [[0] * len2 for _ in range(len1)]

        # calculate shortest edit distance
        for i in range(len1):
            for j in range(len2):
                cost = 0 if word1[i] == word2[j] else 1

                # Compute distances for insertion, deletion and substitution
                insertion = distance[i][j - 1] + 1 if j > 0 else i + 1
                deletion = distance[i - 1][j] + 1 if i > 0 else j + 1
                substitution = distance[i - 1][j - 1] + cost if i > 0 and j > 0 else max(i, j) + cost

                distance[i][j] = min(deletion, insertion, substitution)

                # Compute transposition when relevant
                if i > 0 and j > 0 and word1[i] == word2[j - 1] and word1[i - 1] == word2[j]:
                    transposition = distance[i - 2][j - 2] + 1 if i > 1 and j > 1 else max(i, j) - 1
                    distance[i][j] = min(distance[i][j], transposition)

            # stop early if we surpass cutoff
            if 0 < cutoff <= min(distance[i]):
                return cutoff_return
        return distance[i][j]

    def find_q_gram_matches(self, q_grams: set) -> pd.DataFrame:
        """Find which of the given q_grams exist in self.q_gram_to_word,
        return a sorted dataframe of best matching words.
        """
        n_q_grams = len(q_grams)

        matches = {}

        # find words that match our q-grams
        for q_gram in q_grams:
            if words := self.q_gram_to_word.get(q_gram, False):
                # q_gram exists in our search index
                for word in words:
                    matches[word] = matches.get(word, 0) + words[word]

        # if we find no results, return an empty dataframe
        if len(matches) == 0:
            return pd.DataFrame({"word": [], "matches": []})

        # otherwise, create a dataframe and
        # reduce search results to most relevant results
        matches = {"word": matches.keys(), "matches": matches.values()}
        matches = pd.DataFrame(matches)
        max_q = max(matches["matches"])  # this has the most matching q-grams

        # determine how many results we want to keep based on how good our results are
        min_q = min(max(max_q * 0.32,  # have at least a third of q-grams of best match or...
                        max(n_q_grams * 0.5,  # if more, at least half the q-grams in the query word?
                            1)),  # okay just do 1 q-gram if there are no more in the word
                    max_q)  # never have min_q be over max_q

        matches = matches[matches["matches"] >= min_q]
        matches = matches.sort_values(by="matches", ascending=False)
        matches = matches.reset_index(drop=True)

        return matches.iloc[:min(len(matches), 2500), :]  # return at most this many results

    def spell_check(self, text: str, skip_len=1) -> OrderedDict:
        """Create an OrderedDict of each word in the text (space separated)
        with as values possible alternatives.

        Alternatives are first found with q-grams, then refined with string edit distance

        We rank alternative words based on 1) edit distance 2) how often a word is used in an entry
        If too many results are found, we only keep edit distance 1,
        if we want more results, we keep with longer edit distance up to `never_accept_this`

        word_results = OrderedDict(
            "word": [work]
            )

        NOTE: only ALTERNATIVES are ever returned, this function returns empty list for item BOTH when
            1) the exact word is in the data
            2) when there are no suitable alternatives
        """
        count_occurence = lambda x: sum(self.word_to_identifier[x].values())  # count occurences of a word

        word_results = OrderedDict()

        matches_min = 3  # ideally we have at least this many alternatives
        matches_max = 10  # ideally don't much more than this many matches
        always_accept_this = 1  # values of this edit distance or lower always accepted
        never_accept_this = 4  # values this edit distance or over always rejected

        # make list of unique words
        text = self.clean_text(text)
        words = OrderedDict()
        for word in text.split(" "):
            if len(word) != 0:
                words[word] = False
        words = words.keys()

        for word in words:
            if len(word) <= skip_len:  # dont look for alternatives for text this short
                word_results[word] = []
                continue

            # reduce acceptable edit distance with short words
            dont_accept = int(round(max(1, min((len(word) * 0.66), never_accept_this)), 0))

            # first, find possible matches quickly
            q_grams = self.text_to_positional_q_gram(word)
            possible_matches = self.find_q_gram_matches(set(q_grams))

            first_matches = Counter()
            other_matches = {}

            # now, refine with edit distance
            for row in possible_matches.itertuples():

                edit_distance = self.osa_distance(word, row[1], cutoff=dont_accept)

                if edit_distance == 0:
                    continue  # we are looking for alternatives only, not the exact word
                elif edit_distance <= always_accept_this:
                    first_matches[row[1]] = count_occurence(row[1])
                elif edit_distance < dont_accept:
                    if not other_matches.get(edit_distance):
                        other_matches[edit_distance] = Counter()
                    other_matches[edit_distance][row[1]] = count_occurence(row[1])
                else:
                    continue

            # add matches in correct order:
            matches = [match for match, _ in first_matches.most_common()]
            # if we have fewer matches than goal, add more 'less good' matches
            if len(matches) < matches_min:
                for i in range(always_accept_this + 1, dont_accept):
                    # iteratively increase matches with 'worse' results so we hit goal of minimum alternatives
                    if new := other_matches.get(i):
                        prev_num = 10e100
                        for match, num in new.most_common():
                            if num == prev_num:
                                matches.append(match)
                            elif num != prev_num and len(matches) <= matches_max:
                                matches.append(match)
                            else:
                                break
                            prev_num = num

            word_results[word] = matches
        return word_results

    def build_queries(self, query_text) -> list:
        """Make all possible subsets of words in the query, including alternative words."""
        query_text = self.spell_check(query_text)

        # find all combinations of the query words as given
        queries = list(query_text.keys())
        subsets = list(itertools.chain.from_iterable(
            (itertools.combinations(
                queries, r) for r in range(1, len(queries) + 1))))
        all_queries = []

        for combination in subsets:
            # add the 'default' option
            all_queries.append(combination)
            # now add all options with all alternatives
            for i, word in enumerate(combination):
                for alternative in query_text.get(word, []):
                    alternative_combination = list(combination)
                    alternative_combination[i] = alternative
                    all_queries.append(alternative_combination)

        return all_queries

    def weigh_identifiers(self, identifiers: Counter, weight: int, weighted_ids: Counter) -> Counter:
        """Add weights to identifier counter for these identifiers times how often it occurs in identifier."""
        for identifier, occurrences in identifiers.items():
            weighted_ids[identifier] += (weight * occurrences)
        return weighted_ids

    def search_size_1(self, queries: list, original_words: set, orig_word_weight=5, exact_word_weight=1) -> dict:
        """Return a dict of {query_word: Counter(identifier)}.

        queries: is a list of len 1 tuple/lists of words that are a searched word or a 'spell checked' similar word
        original words: a list of words actually searched for (not including spellchecked)

        orig_word_weight: additional weight to add to original words
        exact_word_weight: additional weight to add to exact word matches (as opposed to be 'in' str)

        First, we find all matching words, creating a dict of words in 'queries' as keys and words matching that query word as list of values
        Next, we convert this to identifiers and add weights:
            Weight will be increased if matching 'orig_word_weight' or 'exact_word_weight'
        """
        matches = {}
        # add each word in search index if query_word in word
        for word in self.word_to_identifier.keys():
            for query in queries:
                # query is list/tuple of len 1
                query_word = query[0]  # only use the word
                if query_word in word:
                    words = matches.get(query_word, [])
                    words.extend([word])
                    matches[query_word] = words

        # now convert matched words to matched identifiers
        matched_identifiers = {}
        for word, matching_words in matches.items():
            for matched_word in matching_words:
                weight = self.base_weight
                id_counter = matched_identifiers.get(word, Counter())

                # add the word n times, where n is the weight, original search word is weighted higher than alternatives
                if matched_word in original_words:
                    weight += orig_word_weight  # increase weight for original word
                if matched_word == word:
                    weight += exact_word_weight  # increase weight for exact matching word

                id_counter = self.weigh_identifiers(self.word_to_identifier[matched_word], weight, id_counter)
                matched_identifiers[word] = id_counter

        return matched_identifiers

    def fuzzy_search(self, text: str, return_counter: bool = False) -> list:
        """Search the dataframe, finding approximate matches and return a list of identifiers,
        ranked by how well each identifier matches the search text.

        1. First, identifiers matching single words (and spell-checked alternatives) are found and weighted.
        2. If the search term consisted of multiple words, combinations of those words are checked next.
            2.1 Increasing in size (first two words, then three etc.), we look for identifiers that contain that set of
            words, these are also weighted, based on the sum of all one-word weights (from first step) and the length
            of the sequence.
            2.2 Next, we also look specifically for combinations occurring next to each other. And add more weight like
            the step above (2.1).
        We multiply the weighting of step 2 by the sequence length, based on the assumption that finding more search
        words will be a more relevant result than just finding a single word, and again if they are in the
        correct order.

        Finally, all found identifiers are sorted on their weight and returned.
        """
        text = text.strip()

        queries = self.build_queries(text)

        # make list of unique original words
        orig_words = OrderedDict()
        for word in text.split(" "):
            orig_words[word] = False
        orig_words = orig_words.keys()
        orig_words = {self.clean_text(word) for word in orig_words}

        # order the queries by the amount of words they contain
        # we do this because longer queries (more words) are harder to find, but we have many alternatives so we search in a smaller search space
        queries_by_size = OrderedDict()
        longest_query = max([len(q) for q in queries])
        for query_len in range(1, longest_query + 1):
            queries_by_size[query_len] = [q for q in queries if len(q) == query_len]

        # first handle queries of length 1
        query_to_identifier = self.search_size_1(queries_by_size[1], orig_words)

        # get all results into a df, we rank further later
        all_identifiers = set()
        for id_list in [id_list for id_list in query_to_identifier.values()]:
            all_identifiers.update(id_list)
        search_df = self.df.loc[list(all_identifiers)]

        # now, we search for combinations of query words and get only those identifiers
        # we then reduce de search_df further for only those matching identifiers
        # we then search the permutations of that set of words
        for q_len, query_set in queries_by_size.items():
            if q_len == 1:
                # we already did these above
                continue
            for query in query_set:
                # get the intersection of all identifiers
                # meaning, a set of identifiers that occur in ALL sets of len(1) for the individual words in the query
                # this ensures we only ever search data where ALL items occur to substantially reduce search-space
                # finally, make this a Counter (with each item=1) so we can properly weigh things later
                query_id_sets = [set(query_to_identifier.get(q_word)) for q_word in query if
                                 query_to_identifier.get(q_word, False)]
                if len(query_id_sets) == 0:
                    continue
                query_identifier_set = set.intersection(*query_id_sets)
                if len(query_identifier_set) == 0:
                    # there is no match for this combination of query words, skip
                    break

                # now we convert the query identifiers to a Counter of 'occurrence',
                # where we weigh queries with only original words higher
                query_identifiers = Counter()
                for identifier in query_identifier_set:
                    weight = 0
                    for query_word in query:
                        # if the query_word and identifier combination exist get score, otherwise 0
                        weight += query_to_identifier.get(query_word, {}).get(identifier, 0)

                    query_identifiers[identifier] = weight

                # we now add these identifiers to a counter for this query name,
                query_name = " ".join(query)

                weight = self.base_weight * q_len
                query_to_identifier[query_name] = self.weigh_identifiers(query_identifiers, weight, Counter())

                # now search for all permutations of this query combined with a space
                query_df = search_df[search_df[self.identifier_name].isin(query_identifiers)]
                for query_perm in itertools.permutations(query):
                    mask = self.filter_dataframe(query_df, " ".join(query_perm), search_columns=["query_col"])
                    new_df = query_df.loc[mask].reset_index(drop=True)
                    if len(new_df) == 0:
                        # there is no match for this permutation of words, skip
                        continue
                    new_id_list = new_df[self.identifier_name]

                    new_ids = Counter()
                    for new_id in new_id_list:
                        new_ids[new_id] = query_identifiers[new_id]

                    # we weigh a combination of words that is next also to each other even higher than just the words separately
                    query_to_identifier[query_name] = self.weigh_identifiers(new_ids, weight,
                                                                             query_to_identifier[query_name])
        # now finally, move to one object sorted list by highest score
        all_identifiers = Counter()
        for identifiers in query_to_identifier.values():
            all_identifiers += identifiers

        if return_counter:
            return all_identifiers
        # now sort on highest weights and make list type
        sorted_identifiers = [identifier for identifier, _ in all_identifiers.most_common()]
        return sorted_identifiers

    def search(self, text) -> list:
        """Search the dataframe on this text, return a sorted list of identifiers."""
        t = time()
        text = text.strip()

        if len(text) == 0:
            logger.debug(f"Empty search, returned all items")
            return self.df.index.to_list()

        fuzzy_identifiers = self.fuzzy_search(text)
        if len(fuzzy_identifiers) == 0:
            logger.debug(f"Found 0 search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
            return []

        # take the fuzzy search sub-set of data and search it literally
        df = self.df.loc[fuzzy_identifiers].copy()

        literal_identifiers = self.literal_search(text, df)
        if len(literal_identifiers) == 0:
            logger.debug(
                f"Found {len(fuzzy_identifiers)} search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
            return fuzzy_identifiers

        # append any fuzzy identifiers that were not found in the literal search
        literal_id_set = set(literal_identifiers)
        remaining_fuzzy_identifiers = [
            _id for _id in fuzzy_identifiers if _id not in literal_id_set]
        identifiers = literal_identifiers + remaining_fuzzy_identifiers

        logger.debug(
            f"Found {len(identifiers)} ({len(literal_identifiers)} literal) search results for '{text}' in {len(self.df)} items in {time() - t:.2f} seconds")
        return identifiers



from typing import TypedDict

from qtpy import QtCore, QtWidgets, QtGui
from qtpy.QtCore import Qt


class CardData(TypedDict):
    title: str
    subtitle: str | None
    detail: str | None
    categories: list[str] | None


class CardDelegate(QtWidgets.QStyledItemDelegate):
    """Delegate for rendering card-like items with title, subtitle, categories and background icon."""

    PADDING = 8
    MARGIN = 2
    TITLE_LINES = 2
    ICON_OPACITY = 0.3

    def sizeHint(self, option, index):
        if index.data() is None:
            return super().sizeHint(option, index)

        # Calculate text heights
        fm = option.fontMetrics
        line_height = fm.height()

        # Title (2 lines, larger font)
        title_height = int(line_height * 1 * self.TITLE_LINES) + 5

        # Subtitle
        subtitle_height = int(line_height * 0.9)  # 0.9x for smaller font

        # Categories
        categories_height = 7 + int(line_height * 0.8)

        # Total height with padding
        total_height = (self.PADDING * 2 + self.MARGIN * 2 +
                       title_height + subtitle_height + categories_height)

        return QtCore.QSize(option.rect.width(), max(total_height, 40))

    def paint(self, painter, option: QtWidgets.QStyleOptionViewItem, index):
        if index.data() is None:
            super().paint(painter, option, index)
            return

        painter.save()
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)

        card_data = index.data()
        is_selected = option.state & QtWidgets.QStyle.StateFlag.State_Selected
        font_size = option.font.pointSize()

        # Draw background and border
        rect = option.rect.adjusted(self.MARGIN, self.MARGIN, -self.MARGIN, -self.MARGIN)

        # Background
        painter.fillRect(rect, option.palette.base())

        # Border
        border_color = option.palette.highlight() if is_selected else option.palette.mid()
        painter.setPen(QtGui.QPen(border_color, 1))
        painter.drawRoundedRect(rect, 3, 3)

        # Draw background icon
        icon = index.data(Qt.ItemDataRole.DecorationRole)
        icon_size = 0
        if icon and not icon.isNull():
            painter.setOpacity(self.ICON_OPACITY)
            icon_size = int(rect.height() * 0.8)
            icon_x = rect.right() - icon_size - 10
            icon_y = rect.top() + (rect.height() - icon_size) // 2
            icon.paint(painter, icon_x, icon_y, icon_size, icon_size)
            painter.setOpacity(1.0)

        # Setup text area
        text_rect = rect.adjusted(self.PADDING, self.PADDING, -self.PADDING, -self.PADDING)
        y = text_rect.top()

        # Draw title (bold, larger, 2 lines)
        title = card_data.get('title', '')
        title_font = option.font
        title_font.setPointSize(int(option.font.pointSize() * 1))
        title_font.setWeight(QtGui.QFont.Weight.DemiBold)
        painter.setFont(title_font)
        painter.setPen(option.palette.text().color())

        title_fm = QtGui.QFontMetrics(title_font)
        title_height = 5 + title_fm.height() * self.TITLE_LINES
        title_rect = QtCore.QRect(text_rect.left(), y, text_rect.width(), title_height)

        # Elide title text if it's too long for 2 lines
        title_text = str(title)
        max_width = title_rect.width()

        # Split into words and fit within 2 lines with eliding
        words = title_text.split()
        line1_words = []
        line2_words = []
        current_line = line1_words

        for word in words:
            test_text = " ".join(current_line + [word])
            if title_fm.horizontalAdvance(test_text) <= max_width:
                current_line.append(word)
            elif current_line is line1_words and len(line2_words) == 0:
                # Move to second line
                current_line = line2_words
                current_line.append(word)
            else:
                # Need to elide
                break

        line1_text = " ".join(line1_words)
        line2_text = " ".join(line2_words)

        # If there are remaining words, elide the second line
        if len(line1_words) + len(line2_words) < len(words):
            line2_text = title_fm.elidedText(title_text if not line1_text else " ".join(words[len(line1_words):]),
                                            Qt.TextElideMode.ElideRight, max_width)

        # Draw title lines
        painter.drawText(title_rect.left(), title_rect.top() + title_fm.ascent(), line1_text)
        if line2_text:
            painter.drawText(title_rect.left(), title_rect.top() + title_fm.ascent() + title_fm.height(), line2_text)

        y += title_height

        # Draw subtitle (smaller)
        subtitle = card_data.get('subtitle', '')
        if subtitle:
            subtitle_font: QtGui.QFont = option.font
            subtitle_font.setPointSize(int(font_size * 0.9))
            subtitle_font.setWeight(QtGui.QFont.Weight.Light)
            painter.setFont(subtitle_font)

            subtitle_fm = QtGui.QFontMetrics(subtitle_font)
            subtitle_height = subtitle_fm.height()
            subtitle_rect = QtCore.QRect(text_rect.left(), y, text_rect.width(), subtitle_height)

            # Elide subtitle if too long
            subtitle_text = subtitle_fm.elidedText(str(subtitle), Qt.TextElideMode.ElideRight, subtitle_rect.width())
            painter.drawText(subtitle_rect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter, subtitle_text)
            y += subtitle_height

        # Draw detail (bottom left)
        detail = card_data.get('detail', '')
        detail_width = 0
        if detail:
            detail_font = option.font
            detail_font.setPointSize(int(font_size * 0.8))
            painter.setFont(detail_font)

            detail_fm = QtGui.QFontMetrics(detail_font)
            detail_height = detail_fm.height()

            # Reserve half width for detail, half for categories
            max_detail_width = text_rect.width() // 2 - 10
            detail_rect = QtCore.QRect(text_rect.left(), text_rect.bottom() - detail_height,
                                       max_detail_width, detail_height)

            # Elide detail if too long
            detail_text_elided = detail_fm.elidedText(str(detail), Qt.TextElideMode.ElideRight, max_detail_width)
            painter.drawText(detail_rect, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter, detail_text_elided)
            detail_width = detail_fm.horizontalAdvance(detail_text_elided) + 10

        # Draw categories (pipe separated, bottom right)
        categories = card_data.get('categories', [])
        if categories and isinstance(categories, (list, tuple)):
            categories_text = "  |  ".join(str(cat) for cat in categories)
            categories_font = option.font
            categories_font.setPointSize(int(font_size * 0.8))
            painter.setFont(categories_font)

            categories_fm = QtGui.QFontMetrics(categories_font)
            categories_height = categories_fm.height()

            # Adjust width to account for detail on left
            available_width = text_rect.width() - detail_width
            categories_rect = QtCore.QRect(text_rect.left() + detail_width, text_rect.bottom() - categories_height,
                                          available_width, categories_height)

            # Elide categories if too long
            categories_text_elided = categories_fm.elidedText(categories_text, Qt.TextElideMode.ElideRight, available_width)
            painter.drawText(categories_rect, Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter, categories_text_elided)

        painter.restore()





# -*- coding: utf-8 -*-

# =============================================================================
# Global Sensitivity Analysis (GSA) functions and class for the Delta
# Moment-Independent measure based on Monte Carlo simulation LCA results.
# see: https://salib.readthedocs.io/en/latest/api.html#delta-moment-independent-measure
# =============================================================================
import os
import traceback
from time import time
from loguru import logger

import bw2calc as bc
import numpy as np
import pandas as pd
import bw2data as bd
from SALib.analyze import delta

# from ..settings import ab_settings
from .montecarlo import MonteCarloLCA, perform_MonteCarlo_LCA

try:
    # attempt bw25 import
    from bw_graph_tools.graph_traversal import NewNodeEachVisitGraphTraversal
except ImportError:
    # standard import on failure
    from bw2calc import GraphTraversal





def get_lca(fu, method):
    """Calculates a non-stochastic LCA and returns a the LCA object."""
    lca = bc.LCA(fu, method=method)
    lca.lci()
    lca.lcia()
    logger.info(f"Non-stochastic LCA score: {lca.score}")

    # add reverse dictionaries
    lca.activity_dict_rev, lca.product_dict_rev, lca.biosphere_dict_rev = (
        lca.reverse_dict()
    )

    return lca


def filter_technosphere_exchanges(lca, cutoff=0.05, max_calc=1000):
    """Use brightway's GraphTraversal to identify the relevant
    technosphere exchanges in a non-stochastic LCA."""
    start = time()
    res = NewNodeEachVisitGraphTraversal.calculate(lca, cutoff=cutoff, max_calc=int(max_calc))

    # get all edges
    technosphere_exchange_indices = []
    for e in res["edges"]:
        if e.consumer_index != -1:  # filter out head introduced in graph traversal
            technosphere_exchange_indices.append((e.producer_index, e.consumer_index))
    logger.info(
        "TECHNOSPHERE {} filtering resulted in {} of {} exchanges and took {} iterations in {} seconds.".format(
            lca.technosphere_matrix.shape,
            len(technosphere_exchange_indices),
            lca.technosphere_matrix.getnnz(),
            res["calculation_count"],
            np.round(time() - start, 2),
        )
    )
    return technosphere_exchange_indices


def filter_biosphere_exchanges(lca, cutoff=0.005):
    """Reduce biosphere exchanges to those that matter for a given impact
    category in a non-stochastic LCA."""
    start = time()

    inv = lca.characterized_inventory
    finv = inv.multiply(abs(inv) > abs(lca.score / (1 / cutoff)))
    biosphere_exchange_indices = list(zip(*finv.nonzero()))
    explained_fraction = finv.sum() / lca.score
    logger.info(
        "BIOSPHERE {} filtering resulted in {} of {} exchanges ({}% of total impact) and took {} seconds.".format(
            inv.shape,
            finv.nnz,
            inv.nnz,
            np.round(explained_fraction * 100, 2),
            np.round(time() - start, 2),
        )
    )
    return biosphere_exchange_indices


def get_exchanges(lca, indices, biosphere=False, only_uncertain=True):
    """Get actual exchange objects from indices.
    By default get only exchanges that have uncertainties.

    Returns
    -------
    exchanges : list
        List of exchange objects
    indices : list of tuples
        List of indices
    """
    exchanges = list()
    for i in indices:
        if biosphere:
            from_act = bd.get_activity(lca.biosphere_dict_rev[i[0]])
        else:  # technosphere
            from_act = bd.get_activity(lca.activity_dict_rev[i[0]])
        to_act = bd.get_activity(lca.activity_dict_rev[i[1]])

        for exc in to_act.exchanges():
            if exc.input == from_act.key:
                exchanges.append(exc)
                # continue  # if there was always only one max exchange between two activities

    # in theory there should be as many exchanges as indices, but since
    # multiple exchanges are possible between two activities, the number of
    # exchanges must be at least equal or higher to the number of indices
    if len(exchanges) < len(
        indices
    ):  # must have at least as many exchanges as indices (assu)
        raise ValueError(
            "Error: mismatch between indices provided ({}) and Exchanges received ({}).".format(
                len(indices), len(exchanges)
            )
        )

    # by default drop exchanges and indices if the have no uncertainties
    if only_uncertain:
        exchanges, indices = drop_no_uncertainty_exchanges(exchanges, indices)

    return exchanges, indices


def drop_no_uncertainty_exchanges(excs, indices):
    excs_no = list()
    indices_no = list()
    for exc, ind in zip(excs, indices):
        if exc.get("uncertainty type") and exc.get("uncertainty type") >= 1:
            excs_no.append(exc)
            indices_no.append(ind)
    logger.info(
        "Dropping {} exchanges of {} with no uncertainty. {} remaining.".format(
            len(excs) - len(excs_no), len(excs), len(excs_no)
        )
    )
    return excs_no, indices_no


def get_exchanges_dataframe(exchanges, indices, biosphere=False):
    """Returns a Dataframe from the exchange data and a bit of additional information."""

    for exc, i in zip(exchanges, indices):
        from_act = bd.get_activity(exc.get("input"))
        to_act = bd.get_activity(exc.get("output"))

        exc.update(
            {
                "index": i,
                "from name": from_act.get("name", np.nan),
                "from location": from_act.get("location", np.nan),
                "to name": to_act.get("name", np.nan),
                "to location": to_act.get("location", np.nan),
            }
        )

        # GSA name (needs to yield unique labels!)
        if biosphere:
            exc.update(
                {
                    "GSA name": "B: {} // {} ({}) [{}]".format(
                        from_act.get("name", ""),
                        to_act.get("name", ""),
                        to_act.get("reference product", ""),
                        to_act.get("location", ""),
                    )
                }
            )
        else:
            exc.update(
                {
                    "GSA name": "T: {} FROM {} [{}] TO {} ({}) [{}]".format(
                        from_act.get("reference product", ""),
                        from_act.get("name", ""),
                        from_act.get("location", ""),
                        to_act.get("name", ""),
                        to_act.get("reference product", ""),
                        to_act.get("location", ""),
                    )
                }
            )

    return pd.DataFrame(exchanges)


def get_CF_dataframe(lca, only_uncertain_CFs=True):
    """Returns a dataframe with the metadata for the characterization factors
    (in the biosphere matrix). Filters non-stochastic CFs if desired (default)."""
    data = dict()
    for params_index, row in enumerate(lca.cf_params):
        if only_uncertain_CFs and row["uncertainty_type"] <= 1:
            continue
        cf_index = row["row"]
        bio_act = bd.get_activity(lca.biosphere_dict_rev[cf_index])

        data.update({params_index: bio_act.as_dict()})

        for name in row.dtype.names:
            data[params_index][name] = row[name]

        data[params_index]["index"] = cf_index
        data[params_index]["GSA name"] = (
            "CF: " + bio_act["name"] + str(bio_act["categories"])
        )

    logger.info(
        "CHARACTERIZATION FACTORS filtering resulted in including {} of {} characteriation factors.".format(
            len(data),
            len(lca.cf_params),
        )
    )
    df = pd.DataFrame(data).T
    df.rename(columns={"uncertainty_type": "uncertainty type"}, inplace=True)
    return df


def get_parameters_DF(mc):
    """Product to make parameters dataframe"""
    if bool(mc.parameter_data):  # returns False if dict is empty
        dfp = pd.DataFrame(mc.parameter_data).T
        dfp["GSA name"] = "P: " + dfp["name"]
        logger.info(f"PARAMETERS: {len(dfp)}")
        return dfp
    else:
        logger.info("PARAMETERS: None included.")
        return pd.DataFrame()  # return emtpy df


def get_exchange_values(matrix, indices):
    """Get technosphere exchanges values from a list of exchanges
    (row and column information)"""
    return [matrix[i] for i in indices]


def get_X(matrix_list, indices):
    """Get the input data to the GSA, i.e. A and B matrix values for each
    model run."""
    X = np.zeros((len(matrix_list), len(indices)))
    for row, M in enumerate(matrix_list):
        X[row, :] = get_exchange_values(M, indices)
    return X


def get_X_CF(mc, dfcf, method):
    """Get the characterization factors used for each model run. Only those CFs
    that are in the dfcf dataframe will be returned (i.e. by default only the
    CFs that have uncertainties."""
    # get all CF inputs
    CF_data = np.array(mc.CF_dict[method])  # has the same shape as the Xa and Xb below

    # reduce this to uncertain CFs only (if this was done for the dfcf)
    params_indices = dfcf.index.values

    # if params_indices:
    return CF_data[:, params_indices]


def get_X_P(dfp):
    """Get the parameter values for each model run"""
    lists = [d for d in dfp["values"]]
    return list(zip(*lists))


def get_problem(X, names):
    return {
        "num_vars": X.shape[1],
        "names": names,
        "bounds": list(zip(*(np.amin(X, axis=0), np.amax(X, axis=0)))),
    }


class GlobalSensitivityAnalysis(object):
    """Class for Global Sensitivity Analysis.
    For now Delta Moment Independent Measure based on:
    https://salib.readthedocs.io/en/latest/api.html#delta-moment-independent-measure
    Builds on top of Monte Carlo Simulation results.
    """

    def __init__(self, mc):
        self.update_mc(mc)
        self.act_number = int()
        self.method_number = int()
        self.cutoff_technosphere = float()
        self.cutoff_biosphere = float()

    def update_mc(self, mc):
        "Update the Monte Carlo Simulation object (and results)."
        try:
            assert isinstance(mc, MonteCarloLCA)
            self.mc = mc
        except AssertionError:
            raise AssertionError(
                "mc should be an instance of MonteCarloLCA, but instead it is a {}.".format(
                    type(mc)
                )
            )

    def perform_GSA(
        self,
        act_number=0,
        method_number=0,
        cutoff_technosphere=0.01,
        cutoff_biosphere=0.01,
    ):
        """Perform GSA for specific reference flow and impact category."""
        start = time()

        # set FU and method
        try:
            self.act_number = act_number
            self.method_number = method_number
            self.cutoff_technosphere = cutoff_technosphere
            self.cutoff_biosphere = cutoff_biosphere

            self.fu = self.mc.cs["inv"][act_number]
            self.activity = bd.get_activity(self.mc.rev_activity_index[act_number])
            self.method = self.mc.cs["ia"][method_number]

        except Exception as e:
            traceback.print_exc()
            # todo: QMessageBox.warning(self, 'Could not perform Delta analysis', str(e))
            logger.error("Initializing the GSA failed.")
            return None

        logger.info(
            f"-- GSA --\n Project: {bd.projects.current} CS: {self.mc.cs_name} "
            f"Activity: {self.activity} Method: {self.method}",
        )

        # get non-stochastic LCA object with reverse dictionaries
        self.lca = get_lca(self.fu, self.method)

        # =============================================================================
        #   Filter exchanges and get metadata DataFrames
        # =============================================================================
        dfs = []
        # technosphere
        if self.mc.include_technosphere:
            self.t_indices = filter_technosphere_exchanges(
                self.lca, cutoff=cutoff_technosphere, max_calc=1e4
            )
            self.t_exchanges, self.t_indices = get_exchanges(self.lca, self.t_indices)
            self.dft = get_exchanges_dataframe(self.t_exchanges, self.t_indices)
            if not self.dft.empty:
                dfs.append(self.dft)

        # biosphere
        if self.mc.include_biosphere:
            self.b_indices = filter_biosphere_exchanges(
                self.lca, cutoff=cutoff_biosphere
            )
            self.b_exchanges, self.b_indices = get_exchanges(
                self.lca, self.b_indices, biosphere=True
            )
            self.dfb = get_exchanges_dataframe(
                self.b_exchanges, self.b_indices, biosphere=True
            )
            if not self.dfb.empty:
                dfs.append(self.dfb)

        # characterization factors
        if self.mc.include_cfs:
            self.dfcf = get_CF_dataframe(
                self.lca, only_uncertain_CFs=True
            )  # None if no stochastic CFs
            if not self.dfcf.empty:
                dfs.append(self.dfcf)

        # parameters
        # todo: if parameters, include df, but remove exchanges from T and B (skipped for now)
        self.dfp = get_parameters_DF(self.mc)  # Empty df if no parameters
        if not self.dfp.empty:
            dfs.append(self.dfp)

        # Join dataframes to get metadata
        self.metadata = pd.concat(dfs, axis=0, ignore_index=True, sort=False)
        self.metadata.set_index("GSA name", inplace=True)

        # =============================================================================
        #     GSA
        # =============================================================================

        # Get X (Technosphere, Biosphere and CF values)
        X_list = list()
        if self.mc.include_technosphere and self.t_indices:
            self.Xa = get_X(self.mc.A_matrices, self.t_indices)
            X_list.append(self.Xa)
        if self.mc.include_biosphere and self.b_indices:
            self.Xb = get_X(self.mc.B_matrices, self.b_indices)
            X_list.append(self.Xb)
        if self.mc.include_cfs and not self.dfcf.empty:
            self.Xc = get_X_CF(self.mc, self.dfcf, self.method)
            X_list.append(self.Xc)
        if self.mc.include_parameters and not self.dfp.empty:
            self.Xp = get_X_P(self.dfp)
            X_list.append(self.Xp)

        self.X = np.concatenate(X_list, axis=1)
        # print('X', self.X.shape)

        # Get Y (LCA scores)
        self.Y = self.mc.get_results_dataframe(act_key=self.activity.key)[
            self.method
        ].to_numpy()

        # log-transformation. This makes it more robust for very uneven distributions of LCA results (e.g. toxicity related impacts).
        # Can only be applied if all Monte-Carlo LCA scores are either positive or negative.
        # Should not be used when LCA scores overlap zero (sometimes positive and sometimes negative)
        # if np.all(self.Y > 0) if self.Y[0] > 0 else np.all(self.Y < 0):  # check if all LCA scores are of the same sign
        #     self.Y = np.log(np.abs(self.Y))  # this makes it more robust for very uneven distributions of LCA results
        if np.all(self.Y > 0):  # all positive numbers
            self.Y = np.log(np.abs(self.Y))
            logger.info("All positive LCA scores. Log-transformation performed.")
        elif np.all(self.Y < 0):  # all negative numbers
            self.Y = -np.log(np.abs(self.Y))
            logger.info("All negative LCA scores. Log-transformation performed.")
        else:  # mixed positive and negative numbers
            logger.warning(
                "Log-transformation cannot be applied as LCA scores overlap zero."
            )

        # print('Filtering took {} seconds'.format(np.round(time() - start, 2)))

        # define problem
        self.names = self.metadata.index  # ['GSA name']
        # print('Names:', len(self.names))
        self.problem = get_problem(self.X, self.names)

        # perform delta analysis
        time_delta = time()
        self.Si = delta.analyze(self.problem, self.X, self.Y, print_to_console=False)
        logger.info(
            "Delta analysis took {} seconds".format(
                np.round(time() - time_delta, 2),
            )
        )

        # put GSA results in to dataframe
        self.dfgsa = pd.DataFrame(self.Si, index=self.names).sort_values(
            by="delta", ascending=False
        )
        self.dfgsa.index.names = ["GSA name"]

        # join with metadata
        self.df_final = self.dfgsa.join(self.metadata, on="GSA name")
        self.df_final.reset_index(inplace=True)
        self.df_final["pedigree"] = [str(x) for x in self.df_final["pedigree"]]

        logger.info("GSA took {} seconds".format(np.round(time() - start, 2)))

    def get_save_name(self):
        save_name = (
            self.mc.cs_name
            + "_"
            + str(self.mc.iterations)
            + "_"
            + self.activity["name"]
            + "_"
            + str(self.method)
            + ".xlsx"
        )
        save_name = save_name.replace(",", "").replace("'", "").replace("/", "")
        return save_name

    def export_GSA_output(self):
        from ..settings import ab_settings

        save_name = "gsa_output_" + self.get_save_name()
        self.df_final.to_excel(os.path.join(ab_settings.data_dir, save_name))

    def export_GSA_input(self):
        """Export the input data to the GSA with a human readible index"""
        from ..settings import ab_settings

        X_with_index = pd.DataFrame(self.X.T, index=self.metadata.index)
        save_name = "gsa_input_" + self.get_save_name()
        X_with_index.to_excel(os.path.join(ab_settings.data_dir, save_name))


if __name__ == "__main__":
    mc = perform_MonteCarlo_LCA(project="ei34", cs_name="kraft paper", iterations=20)
    g = GlobalSensitivityAnalysis(mc)
    g.perform_GSA(
        act_number=0, method_number=1, cutoff_technosphere=0.01, cutoff_biosphere=0.01
    )
    g.export()



from typing import Any

from activity_browser import app
from activity_browser.app.actions.base import ABAction, exception_dialogs
from bw2data import get_activity
from bw2data.parameters import (ActivityParameter, Group,
                                                     GroupDependency,
                                                     parameters)
from activity_browser.ui.icons import qicons
from activity_browser.bwutils.utils import Parameter


class ParameterDelete(ABAction):
    """
    ABAction to delete an existing parameter.
    """

    icon = qicons.delete
    text = "Delete parameter..."

    @staticmethod
    @exception_dialogs
    def run(parameter: Any or list[Any]):
        if not isinstance(parameter, list):
            parameter = [parameter]

        for parameter in parameter:
            if isinstance(parameter, Parameter):
                parameter = parameter.to_peewee_model()

            if isinstance(parameter, ActivityParameter):
                db = parameter.database
                code = parameter.code
                amount = (
                    ActivityParameter.select()
                    .where(
                        (ActivityParameter.database == db)
                        & (ActivityParameter.code == code)
                    )
                    .count()
                )

                if amount > 1:
                    parameter.delete_instance()
                else:
                    group = parameter.group
                    act = get_activity((db, code))
                    parameters.remove_from_group(group, act)
                    # Also clear the group if there are no more parameters in it

                    if (
                        not ActivityParameter.select()
                        .where(ActivityParameter.group == group)
                        .exists()
                    ):
                        Group.delete().where(Group.name == group).execute()
                        GroupDependency.delete().where(
                            GroupDependency.group == group
                        ).execute()
            else:
                parameter.delete_instance()
        # After deleting things, recalculate and signal changes
        parameters.recalculate()

        # No fire when everything is still fresh after recalculation, so need to fire manually to be sure everything is
        # updated correctly.
        app.signals.parameter.recalculated.emit()



from importlib.metadata import PackageNotFoundError, version

# get AB version
try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "0.0.0"

# supported EI versions
__ei_versions__ = ["3.4", "3.5", "3.6", "3.7", "3.7.1", "3.8", "3.9", "3.9.1"]



from qtpy import QtWidgets

from .buttons import ABCloseButton, ABMinimizeButton


class ABTabWidget(QtWidgets.QTabWidget):
    def __init__(self, *args, **kwargs):
        """
        Initialize the GroupTabWidget.

        Args:
            name (str): The name of the group, used as the object name for the widget.
            *args: Additional positional arguments passed to the parent QTabWidget.
        """
        super().__init__(*args, **kwargs)
        self.setMovable(True)  # Allow tabs to be rearranged.
        self.setTabsClosable(True)  # Allow tabs to be closed.
        self.tabBar().setExpanding(False)
        
    
    def resizeEvent(self, event):
        super().resizeEvent(event)
        # Force the tab bar to always fill the full width
        self.tabBar().setMinimumWidth(self.width())

    def addTab(self, widget, label, show_minimize=False):
        """Override addTab to add custom buttons to each tab.
        
        Args:
            widget: The widget to add as a tab
            label: The label for the tab
            show_minimize: If True, show minimize button; if False, show close button
        """
        index = super().addTab(widget, label)
        self._set_buttons(index, widget, show_minimize)
        return index
    
    def insertTab(self, index, widget, label, show_minimize=False):
        """Override insertTab to add custom buttons to each tab.
        
        Args:
            index: The index at which to insert the tab
            widget: The widget to add as a tab
            label: The label for the tab
            show_minimize: If True, show minimize button; if False, show close button
        """
        index = super().insertTab(index, widget, label)
        self._set_buttons(index, widget, show_minimize)
        return index
    
    def _set_buttons(self, index, widget, show_minimize=False):
        tab_bar = self.tabBar()
        button = ABMinimizeButton() if show_minimize else ABCloseButton()
        tab_bar.setTabButton(index, QtWidgets.QTabBar.ButtonPosition.RightSide, button)
        button.clicked.connect(lambda w=widget: self.closeTabByWidget(w))
    
    def closeTabByWidget(self, widget):
        """Handle close button click using the widget reference."""
        index = self.indexOf(widget)
        if index >= 0:
            self.tabCloseRequested.emit(index)



import platformdirs
from pathlib import Path

import bw2data as bd


def get_package_path() -> Path:
    path = Path(__file__).resolve().parents[1]
    path.mkdir(parents=True, exist_ok=True)
    return path

def get_appdata_path() -> Path:
    path = Path(platformdirs.user_data_dir(appname="ActivityBrowser", appauthor="pylca"))
    path.mkdir(parents=True, exist_ok=True)
    return path

def get_project_path() -> Path:
    path =  bd.projects.dir
    path.mkdir(parents=True, exist_ok=True)
    return path

def get_project_ab_path() -> Path:
    path =  Path(bd.projects.dir) / "activity_browser"
    path.mkdir(parents=True, exist_ok=True)
    return path



from typing import Literal

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt


class ABDropOverlay(QtWidgets.QWidget):
    opacityMap = {
        "low": 100,
        "medium": 150,
        "high": 200,
    }

    def __init__(self, parent=None, text="Drop here to create new exchanges"):
        super().__init__(parent)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        self.setAttribute(Qt.WidgetAttribute.WA_NoSystemBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAutoFillBackground(False)
        self.resize(parent.size())

        self._text = text
        self._opacity: Literal["low", "medium", "high"] = "medium"

    def hovering(self) -> bool:
        cursor_pos = QtGui.QCursor.pos()
        widget_rect = self.rect()
        local_pos = self.mapFromGlobal(cursor_pos)
        return widget_rect.contains(local_pos)

    def setOpacity(self, level: Literal["low", "medium", "high"]):
        if level in self.opacityMap:
            self._opacity = level
            self.update()

    def opacity(self):
        return self._opacity

    def text(self):
        return self._text

    def setText(self, text: str):
        self._text = text
        self.update()

    def showEvent(self, event):
        self.resize(self.parent().size())
        super().showEvent(event)

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.fillRect(self.rect(), QtGui.QColor(0, 100, 255, self.opacityMap[self.opacity()]))  # Semi-transparent blue
        painter.setPen(Qt.GlobalColor.white)

        font = self.font()
        font.setBold(True)

        painter.setFont(font)
        painter.drawText(self.rect(), Qt.AlignmentFlag.AlignCenter, self.text())



from qtpy import QtWidgets
from loguru import logger

from activity_browser.ui import widgets, delegates, core
from activity_browser.app import metadata, signals


class MetaDataStorePage(QtWidgets.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("MetaDataStorePage")

        self.model = core.ABTreeModel(metadata.dataframe, self, chunk_size=50)
        self.view = MDSView(self)
        self.view.setModel(self.model)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        signals.metadata.synced.connect(self.sync)

    def sync(self):
        logger.log("SYNC", f"{self.__class__.__name__}: {id(self)}")
        self.model.set_dataframe(metadata.dataframe)

    def build_layout(self):
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        self.setLayout(layout)


class MDSView(widgets.ABTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setItemDelegate(delegates.StringDelegate(self))



from qtpy import QtWidgets
from typing import Callable
from inspect import signature


class ABMenu(QtWidgets.QMenu):
    menuSetup: list[Callable[["ABMenu", QtWidgets.QWidget], None]]
    title: str = None

    def __init__(self, pos=None, parent=None, title: str = None):
        super().__init__(parent)

        for item in self.menuSetup:
            if len(signature(item).parameters) == 1:
                item(self)
            if len(signature(item).parameters) == 2:
                item(self, parent)

    def add(self, action, *args, enable=True, text=None, **kwargs):
        qaction = action.get_QAction(*args, parent=self, enabled=enable, text=text, **kwargs)
        self.addAction(qaction)

    def callback(self, text: str, func: Callable, args: list = None, kwargs: dict = None):
        args = args or []
        kwargs = kwargs or {}

        action = QtWidgets.QAction(text, self)
        action.triggered.connect(lambda: func(*args, **kwargs))
        self.addAction(action)



from typing import Any, List

import bw2data as bd

from activity_browser import app
from activity_browser.app.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.dialogs import UncertaintyDialog


class ExchangeUncertaintyModify(ABAction):
    """
    ABAction to open the UncertaintyWizard for an exchange
    """

    icon = qicons.edit
    text = "Modify uncertainty"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[bd.Edge]):

        ok, uc_dict = UncertaintyDialog.get_uncertainty_dict(
            parent=app.main_window,
            initial=exchanges[0].uncertainty,
            )
        
        if not ok:
            return
        
        for exchange in exchanges:
            for key, value in uc_dict.items():
                exchange[key] = value
            exchange.save()



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets
from stats_arrays import uncertainty_choices as uc

from activity_browser.ui.dialogs import UncertaintyDialog


class UncertaintyDelegate(QtWidgets.QStyledItemDelegate):
    """A combobox containing the sorted list of possible uncertainties
    `setModelData` stores the integer id of the selected uncertainty
    distribution.
    """
    def displayText(self, value, locale):
        """Take the given integer id and return the description.

        Will return the 'Unknown' uncertainty description if the given id
        either cannot be found or the value is 'nan' (when id is not set)
        """
        if isinstance(value, (int, float)) and int(value) in uc.id_dict:
            return uc.id_dict[int(value)].description
        elif isinstance(value, dict) and value.get("uncertainty type") in uc.id_dict:
            return uc[value["uncertainty type"]].description
        return uc[0].description

    def createEditor(self, parent, option, index):
        """Simply use the wizard for updating uncertainties. Send a signal."""
        from activity_browser import app

        item = index.internalPointer()
        item_name = item.__class__.__name__

        if item_name == "ParametersItem" or item_name == "ProjectParametersItem":
            app.actions.ParameterUncertaintyModify.run(item["_parameter"].to_peewee_model())
        elif item_name == "ExchangesItem":
            app.actions.ExchangeUncertaintyModify.run([item.exchange])
        elif item_name == "CharacterizationFactorsItem":
            app.actions.CFUncertaintyModify.run(
                item["_impact_category_name"], [(item["_id"], item["_cf"]),]
            )
        elif isinstance(index.data(), dict):
            return UncertaintyDialog(parent=app.main_window, initial=index.data())

    def setEditorData(self, editor, index: QtCore.QModelIndex):
        pass

    def updateEditorGeometry(self, editor, option, index):
        pass

    def setModelData(
        self,
        editor: UncertaintyDialog,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Read the current text and look up the actual ID of that uncertainty type."""
        if not editor.result() == QtWidgets.QDialog.Accepted:
            return
        
        model.setData(index, editor.result_dict, QtCore.Qt.EditRole)
